#!/usr/bin/env python
"""
Reformat Epstein text message files for readability and count email senders.

    Run: 'EPSTEIN_DOCS_DIR=/path/to/TXT epstein_generate'
"""
import re
from sys import exit

from dotenv import load_dotenv
load_dotenv()
from rich.markup import escape
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from epstein_files.epstein_files import EpsteinFiles, document_cls
from epstein_files.documents.document import INFO_PADDING, Document
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import Email
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile
from epstein_files.util.constant.output_files import make_clean
from epstein_files.util.constant.strings import HOUSE_OVERSIGHT_NOV_2025_ID_REGEX
from epstein_files.util.data import flatten
from epstein_files.util.env import args
from epstein_files.util.file_helper import coerce_file_path, extract_file_id
from epstein_files.util.logging import exit_with_error, logger
from epstein_files.util.output import (print_doj_files, print_emails_section, print_json_files, print_stats,
     print_other_files_section, print_text_messages_section, print_email_timeline, print_emailers_info,
     print_json_metadata, write_urls)
from epstein_files.util.rich import (build_highlighter, console, highlighter, print_color_key, print_json,
     print_title_page_header, print_title_page_tables, print_subtitle_panel, write_html)
from epstein_files.util.timer import Timer
from epstein_files.util.word_count import write_word_counts_html


def generate_html() -> None:
    if args.make_clean:
        make_clean()
        write_urls()
        exit()

    timer = Timer()
    epstein_files = EpsteinFiles.get_files(timer)

    if args.emailers_info:
        print_emailers_info(epstein_files)
        exit()
    elif args.json_metadata:
        print_json_metadata(epstein_files)
        exit()
    elif args.json_files:
        print_json_files(epstein_files)
        exit()

    print_title_page_header()

    if args.email_timeline:
        print_color_key()
    else:
        print_title_page_tables(epstein_files)

    if args.colors_only:
        exit()

    if args.output_doj_files:
        printed_doj_files = print_doj_files(epstein_files)
        timer.log_section_complete('DojFile', epstein_files.doj_files, printed_doj_files)

    if args.output_texts:
        printed_logs = print_text_messages_section(epstein_files)
        timer.log_section_complete('MessengerLog', epstein_files.imessage_logs, printed_logs)

    if args.output_emails:
        printed_emails = print_emails_section(epstein_files)
        timer.log_section_complete('Email', epstein_files.emails, printed_emails)
    elif args.email_timeline:
        print_email_timeline(epstein_files)
        timer.print_at_checkpoint(f"Printed chronological emails table")

    if args.output_other:
        printed_files = print_other_files_section(epstein_files)
        timer.log_section_complete('OtherFile', epstein_files.other_files, printed_files)

    write_html(args.build)
    logger.warning(f"Total time: {timer.seconds_since_start_str()}")

    if args.debug:
        highlighter.print_highlight_counts(console)

    if args.stats:
        print_stats(epstein_files)  # Used for building pytest checks


def epstein_diff():
    """Diff the cleaned up text of two files."""
    Document.diff_files(args.positional_args)


def epstein_grep():
    """Search the cleaned up text of the files."""
    epstein_files = EpsteinFiles.get_files()

    if HOUSE_OVERSIGHT_NOV_2025_ID_REGEX.match(args.positional_args[0]):
        logger.warning(f"'{args.positional_args[0]}' seems to be an ID, running epstein_show instead...")
        epstein_show()
        return

    for search_term in args.positional_args:
        temp_highlighter = build_highlighter(search_term)
        search_results = epstein_files.docs_matching(search_term, args.names)
        print_subtitle_panel(f"Found {len(search_results)} documents matching '{search_term}'")
        last_document = None

        for search_result in search_results:
            doc = search_result.document
            lines = search_result.lines

            if (isinstance(doc, Email) and not args.output_emails) \
                    or (isinstance(doc, (DojFile, OtherFile)) and not args.output_other) \
                    or (isinstance(doc, MessengerLog) and not args.output_texts):
                doc.log(f"{type(doc).__name__} Skipping search result...")
                continue
            elif isinstance(doc, Email) and args.email_body:
                lines = [l for l in search_result.lines if l.line_number > doc.header.num_header_rows]

                if not lines:
                    doc.log(f"None of the matches for '{search_term}' seem to be in the body of the email")
                    continue

            if doc.is_duplicate:
                if last_document and not last_document.is_duplicate:
                    console.line()

                last_document = doc
                console.print(doc.duplicate_file_txt)
            elif args.whole_file:
                console.print(doc)
            else:
                console.print(doc.summary_panel)

                for matching_line in lines:
                    line_txt = matching_line.__rich__()
                    console.print(Padding(temp_highlighter(line_txt), INFO_PADDING), style='gray37')

            console.line()
            console.print(doc.local_path_and_url + '\n', style='dim')


def epstein_show():
    """Show the color highlighted file. If --raw arg is passed, show the raw text of the file as well."""
    raw_docs: list[Document] = []
    console.line()

    try:
        if args.names:
            people = EpsteinFiles.get_files().person_objs(args.names)
            raw_docs = [doc for doc in flatten([p.emails for p in people])]
        else:
            ids = [extract_file_id(arg.strip().strip('_')) for arg in args.positional_args]
            logger.info(f"extracted IDs: {ids}")
            raw_docs = [Document.from_file_id(id) for id in ids]
            logger.info(f"raw docs: {raw_docs}")

        # Rebuild the Document objs so we can see result of latest processing
        docs = Document.sort_by_timestamp([document_cls(doc)(doc.file_path) for doc in raw_docs])
        logger.info(f"Document types: {[doc._class_name for doc in docs]}")
    except Exception as e:
        console.print_exception()
        exit_with_error(str(e))

    for doc in docs:
        console.print('\n', doc, '\n')

        if args.raw:
            console.print(Panel(Text("RAW: ").append(doc.summary()), expand=False, style=doc.border_style))
            console.print(escape(doc.raw_text()), '\n')

            if isinstance(doc, Email):
                console.print(Panel(Text("actual_text: ").append(doc.summary()), expand=False, style=doc.border_style))
                console.print(escape(doc._extract_actual_text()), '\n')
                metadata = doc.metadata
                metadata['is_fwded_article'] = doc.is_fwded_article
                metadata['is_word_count_worthy'] = doc.is_word_count_worthy
                metadata['_is_first_for_user'] = doc._is_first_for_user
                print_json(f"{doc.file_id} Metadata", metadata)

        console.print(doc.local_path_and_url, style='dim')


def epstein_word_count() -> None:
    write_word_counts_html()



import json
import re
from copy import deepcopy
from dataclasses import Field, asdict, dataclass, field, fields
from datetime import datetime
from typing import Generator, Literal

from dateutil.parser import parse

from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.data import remove_zero_time, without_falsey
from epstein_files.util.env import args

DuplicateType = Literal['bounced', 'earlier', 'quoted', 'redacted', 'same']
Metadata = dict[str, bool | datetime | int | str | list[str | None] |dict[str, bool | str]]

# Misc
INDENT = '    '
INDENT_NEWLINE = f'\n{INDENT}'
INDENTED_JOIN = f',{INDENT_NEWLINE}'
MAX_LINE_LENGTH = 135
REPUTATION_MGMT = f'{REPUTATION} management'
SAME = 'same'

DUPE_TYPE_STRS: dict[DuplicateType, str] = {
    'bounced': 'a bounced copy of',
    'earlier': 'an earlier draft of',
    'quoted': 'quoted in full in',
    'redacted': 'a redacted version of',
    SAME: 'the same as',
}

FIELD_SORT_KEY = {
    'id': 'a',
    'author': 'aa',
    'comment': 'zz',
    'duplicate_ids': 'dup',
    'duplicate_of_id': 'dupe',
    'recipients': 'aaa',
}

FINANCIAL_REPORTS_AUTHORS = [
    BOFA_MERRILL,
    DEUTSCHE_BANK,
    ELECTRON_CAPITAL_PARTNERS,
    GOLDMAN_INVESTMENT_MGMT,
    'Invesco',
    JP_MORGAN,
    'Morgan Stanley',
    'S&P',
]

# Fields like timestamp and author are better added from the Document object
NON_METADATA_FIELDS = [
    'actual_text',
    'id',
    'is_synthetic',
]


@dataclass(kw_only=True)
class DocCfg:
    """
    Encapsulates info about files that needs to be manually configured because it cannot be programmatically inferred.

    Attributes:
        id (str): ID of file
        author (Name): Author of the document (if any)
        category (str | None): Type of file
        date (str | None): Parsed to a datetime by timestamp() if it exists
        dupe_type (DuplicateType | None): The type of duplicate this file is or its 'duplicate_ids' are
        duplicate_ids (list[str]): IDs of *other* documents that are dupes of this document
        duplicate_of_id (str | None): If this is a dupe the ID of the duplicated file. This file will be suppressed
        is_interesting (bool | None): Override other considerations and always consider this file interesting (or not)
        is_synthetic (bool): True if this config was generated by the duplicate_cfgs() method
        replace_text_with (bool): True if `description` should replace body of the document when printing.
    """
    id: str
    attached_to_email_id: str | None = None
    author: Name = None
    category: str | None = None
    comment: str = ''
    date: str | None = None
    description: str | None = None
    dupe_type: DuplicateType | None = None
    duplicate_ids: list[str] = field(default_factory=list)
    duplicate_of_id: str | None = None
    is_attribution_uncertain: bool = False
    is_interesting: bool | None = None
    is_synthetic: bool = False
    replace_text_with: str = ''

    @property
    def complete_description(self) -> str | None:
        """String that summarizes what is known about this document."""
        description = ''

        if self.category and not self.description and not self.author:
            if self.category == JUNK:
                return None
            else:
                description = self.category
        elif self.category == REPUTATION:
            author_str = f"{self.author} " if self.author else ''
            description = f"{REPUTATION_MGMT}: {author_str}{self.description}"
        elif self.category == SKYPE_LOG:
            msg = f"{self.category} of conversation with {self.author}" if self.author else self.category
            description = f"{msg} {self.description}" if self.description else msg
        elif self.author and self.description:
            if self.category in [ACADEMIA, BOOK]:
                title = self.description if '"' in self.description else f'"{self.description}"'
                description = f"{title} by {self.author}"
            elif self.category == FINANCE and self.author in FINANCIAL_REPORTS_AUTHORS:
                description = f'{self.author} report: "{self.description}"'
            elif self.category == LEGAL and 'v.' in self.author:
                description = f"{self.author}: {self.description}"

        if not description:
            pieces = without_falsey([self.author, self.description])

            if pieces:
                description = ' '.join(pieces)
            else:
                return None

        if self.attached_to_email_id:
            description += f" attached to email {self.attached_to_email_id}"

        return description

    @property
    def metadata(self) -> Metadata:
        metadata = {k: v for k, v in asdict(self).items() if k not in NON_METADATA_FIELDS and v}

        if self.is_interesting is False:
            metadata['is_interesting'] = False

        return metadata

    @property
    def timestamp(self) -> datetime | None:
        if self.date:
            return parse(self.date)

    def __post_init__(self):
        if self.duplicate_of_id or self.duplicate_ids:
            self.dupe_type = self.dupe_type or SAME

    def duplicate_cfgs(self) -> Generator['DocCfg', None, None]:
        """Create synthetic DocCfg objects that set the 'duplicate_of_id' field to point back to this object."""
        for id in self.duplicate_ids:
            dupe_cfg = deepcopy(self)
            dupe_cfg.id = id
            dupe_cfg.duplicate_of_id = self.id
            dupe_cfg.duplicate_ids = []
            dupe_cfg.dupe_type = self.dupe_type
            dupe_cfg.is_synthetic = True
            yield dupe_cfg

    def _props_strs(self) -> list[str]:
        props = []
        add_prop = lambda f, value: props.append(f"{f.name}={value}")

        for _field in sorted(fields(self), key=lambda f: FIELD_SORT_KEY.get(f.name, f.name)):
            value = getattr(self, _field.name)

            if _field.name in ['actual_text', 'is_fwded_article', 'is_interesting']:  # fields can be False or None or ''
                if value is not None:
                    add_prop(_field, str(value))
            elif not value or _field.name == 'dupe_type' and value == 'same':
                continue
            elif _field.name == AUTHOR:
                add_prop(_field, constantize_name(str(value)) if args.constantize else f"'{value}'")
            elif _field.name == 'recipients':
                recipients_str = str([constantize_name(r) if (args.constantize and r) else r for r in value])
                add_prop(_field, recipients_str.replace("'", '') if args.constantize else recipients_str)
            elif isinstance(value, str):
                if "'" in value:
                    value = '"' + value.replace('"', r'\"') + '"'
                else:
                    value = "'" + value.replace("'", r'\'') + "'"

                add_prop(_field, value)
            else:
                add_prop(_field, str(value))

        return props

    def __repr__(self) -> str:
        props = self._props_strs()
        type_str = f"{type(self).__name__}("
        single_line_repr = type_str + ', '.join(props) + f')'

        if len(single_line_repr) < MAX_LINE_LENGTH or (self.comment and getattr(self, 'is_fwded_article')):
            repr_str = single_line_repr
        else:
            repr_str = f"{type_str}{INDENT_NEWLINE}" + INDENTED_JOIN.join(props)
            repr_str += ',' if props else ''
            repr_str += '\n)'

        if args.constantize:
            repr_str = INDENT + INDENT_NEWLINE.join(repr_str.split('\n'))
            return repr_str.replace(',,', ',').replace(',),', '),').replace(',),', '),')
        else:
            return repr_str


@dataclass(kw_only=True)
class CommunicationCfg(DocCfg):
    """
    Manual config is always required for MessengerLog author attribution. It's also often needed for Email
    files to handle the terrible OCR text that Congress provided which messes up a lot of the email headers.

    Attributes:
        attribution_reason (str | None): Optional explanation of why this email was attributed to this author.
        is_attribution_uncertain (bool): True if we have a good idea of who the author is but are not 100% certain
    """
    attribution_reason: str | None = None

    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(kw_only=True)
class EmailCfg(CommunicationCfg):
    """
    Attributes:
        actual_text (str | None): In dire cases of broken OCR we just configure the body of the email as a string.
        fwded_text_after (str | None): If set, any text after this is a fwd of an article or similar
        is_fwded_article (bool): True if this is a newspaper article someone fwded. Used to exclude articles from word counting.
        recipients (list[Name]): Who received the email
        subject (str): Subject line
    """
    actual_text: str | None = None
    fwded_text_after: str | None = None
    is_fwded_article: bool | None = None
    recipients: list[Name] = field(default_factory=list)
    subject: str | None = None
    truncate_to: int | None = None

    # This is necessary because for some dumb reason @dataclass(repr=False) doesn't cut it
    def __repr__(self) -> str:
        return super().__repr__()


@dataclass(kw_only=True)
class TextCfg(CommunicationCfg):
    # This is necessary because for some dumb reason @dataclass(repr=False) doesn't cut it
    def __repr__(self) -> str:
        return super().__repr__()



from rich.table import Table
from rich.text import Text

from epstein_files.util.rich import highlighter

HEADER_INDENT = Text('        ')
VERTICAL_BAR = '┃ '  # ⎹┃┇┋❘⦀🁢⏐┃⎹
TOP_BAR = '🁢 '


class LeftBarPanel(Table):
    """Create a faux `Panel` that just has a single vertical line down the left side."""
    @classmethod
    def build(cls, text: str | Text, bar_style: str, header: str | Text = ''):
        table = cls.grid(padding=0)
        table.add_column(justify='left', style=bar_style)  # Column for the line
        table.add_column(justify='left')                   # Column for content

        if header:
            table.add_row(TOP_BAR, header)
            table.add_row(VERTICAL_BAR, '')

        for txt_line in highlighter(text).split('\n'):
            table.add_row(VERTICAL_BAR, txt_line)

        return table



import re
from dataclasses import dataclass, field
from datetime import datetime
from typing import ClassVar, Self

from rich.console import Console, ConsoleOptions, RenderableType, RenderResult
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from epstein_files.documents.document import INFO_INDENT, Document
from epstein_files.documents.email import Email
from epstein_files.documents.emails.email_header import FIELDS_COLON_PATTERN
from epstein_files.documents.other_file import Metadata, OtherFile
from epstein_files.util.constant.names import RENATA_BOLOTOVA
from epstein_files.util.constants import FALLBACK_TIMESTAMP
from epstein_files.util.data import without_falsey
from epstein_files.util.layout.left_bar_panel import LeftBarPanel
from epstein_files.util.logging import logger
from epstein_files.util.rich import RAINBOW, INFO_STYLE, SKIPPED_FILE_MSG_PADDING, highlighter, link_text_obj

CHECK_LINK_FOR_DETAILS = 'not shown here, check original PDF for details'
IMAGE_PANEL_REGEX = re.compile(r"\n╭─* Page \d+, Image \d+.*?╯\n", re.DOTALL)
IGNORE_LINE_REGEX = re.compile(r"^(\d+\n?|[\s+❑]{2,})$")
MIN_VALID_LENGTH = 10

OTHER_DOC_URLS = {
    '245-22.pdf': 'https://www.justice.gov/multimedia/Court%20Records/Government%20of%20the%20United%20States%20Virgin%20Islands%20v.%20JPMorgan%20Chase%20Bank,%20N.A.,%20No.%20122-cv-10904%20(S.D.N.Y.%202022)/245-22.pdf'
}

# DojFile specific repair
OCR_REPAIRS: dict[str | re.Pattern, str] = {
    re.compile(fr"({FIELDS_COLON_PATTERN}.*\n)\nSubject:", re.MULTILINE): r'\1Subject:',
}

BAD_DOJ_FILE_IDS = [
    'EFTA00008511',
    'EFTA00008503',
    'EFTA00002512',
    'EFTA00008501',
    'EFTA00008500',
    'EFTA00008514',
    'EFTA00001940',
    'EFTA00008410',
    'EFTA00008411',
    'EFTA00008519',
    'EFTA00008493',
    'EFTA00008527',
    'EFTA00008473',
    'EFTA00001846',
    'EFTA00000052',
    'EFTA00008445',
    'EFTA00008480',
    'EFTA00001124',
    'EFTA00002509',
    'EFTA00008497',
    'EFTA00001031',
    'EFTA00005495',
    'EFTA00002830',
    'EFTA00001937',
    'EFTA00008496',
    'EFTA00008441',
    'EFTA00008415',
    'EFTA00000675',
    'EFTA00002538',
    'EFTA00000672',
    'EFTA00002814',
    'EFTA00002812',
    'EFTA00002543',
    'EFTA00002813',
    'EFTA00002523',
    'EFTA00002079',
    'EFTA00002805',
    'EFTA00001840',
    'EFTA00001114',
    'EFTA00002812',
    'EFTA00002543',
    'EFTA00002786',
    'EFTA00001271',
    'EFTA00002523',
    'EFTA00001979',
    'EFTA00002110',
    'EFTA00008504',
    'EFTA00001368',
    'EFTA00000134',
    'EFTA00000471',
    'EFTA00001848',
    'EFTA00008506',
]

PHONE_BILL_IDS = {
    'EFTA00006770': 'covering 2006-02-01 to 2006-06-16',
    'EFTA00006870': 'covering 2006-02-09 to 2006-07',
    'EFTA00006970': 'covering 2006-04-15 to 2006-07-16',
    # 'EFTA00007070':  # TODO: not a messy phone bill, short, has additional info at end
}

INTERESTING_DOJ_FILES = {
    'EFTA02640711': 'Jabor Y home address (HBJ)',
    'EFTA00039689': 'Dilorio emails to SEC about Signature Bank, Hapoalim, Bioptix / RIOT, Honig, etc.',
    'EFTA00039025': 'Investigation and Review of the Federal Bureau of Prisons Custody, Care, and Supervision of Jeffrey Epstein',
    'EFTA02296929': f"{RENATA_BOLOTOVA} appears to know Epstein's final girlfriend",
    'EFTA01273102': f"payment from Epstein to {RENATA_BOLOTOVA}'s father's account at Sberbank",
}

NO_IMAGE_SUFFIX = """
╭──── Page 1, Image 1 ─────╮
│ (no text found in image) │
╰──────────────────────────╯
""".strip()


@dataclass
class DojFile(OtherFile):
    """
    Class for the files released by DOJ on 2026-01-30 with `EFTA000` prefix.
    """
    _border_style: str | None = None

    border_style_rainbow_idx: ClassVar[int] = 0  # ClassVar to help change color as we print, no impact beyond fancier output
    max_timestamp: ClassVar[datetime] = datetime(2025, 1, 29) # Overloaded in DojFile

    @property
    def border_style(self) -> str:
        """Use a rainbow to make sure each printed object has different color for those before and after."""
        if self._border_style is None:
            self._border_style = RAINBOW[int(self.border_style_rainbow_idx % len(RAINBOW))]
            type(self).border_style_rainbow_idx += 1

        return self._border_style

    @property
    def info(self) -> list[Text]:
        """Overloads superclass to adjust formatting."""
        return [Text(' ').append(sentence) for sentence in super().info]

    @property
    def is_bad_ocr(self) -> bool:
        return self.file_id in BAD_DOJ_FILE_IDS

    @property
    def is_empty(self) -> bool:
        """Overloads superclass method."""
        return len(self.text.strip().removesuffix(NO_IMAGE_SUFFIX)) < MIN_VALID_LENGTH

    @property
    def prettified_text(self) -> Text:
        """Returns the string we want to print as the body of the document."""
        style = ''

        if self.file_id in PHONE_BILL_IDS:
            pages = self.text.split('MetroPCS')
            text = f"{pages[0]}\n\n(Redacted phone bill {PHONE_BILL_IDS[self.file_id]} {CHECK_LINK_FOR_DETAILS})"
        elif self.config and self.config.replace_text_with:
            if len(self.config.replace_text_with) < 300:
                style = INFO_STYLE
                text = f'(Text of {self.config.replace_text_with} {CHECK_LINK_FOR_DETAILS})'
            else:
                text = self.config.replace_text_with
        else:
            text = self.text

        return Text(text, style)

    @property
    def timestamp_sort_key(self) -> tuple[datetime, str, int]:
        """Overloads parent method."""
        dupe_idx = 0
        # TODO: Years of 2001 are often garbage pared from '1.6' etc.
        sort_timestamp = self.timestamp or FALLBACK_TIMESTAMP
        sort_timestamp = FALLBACK_TIMESTAMP if sort_timestamp.year <= 2001 else sort_timestamp
        return (sort_timestamp, self.file_id, dupe_idx)

    def __post_init__(self):
        super().__post_init__()

        if self.file_id in PHONE_BILL_IDS:
            self.strip_image_ocr_panels()

    def doj_link(self) -> Text:
        """Link to this file on the DOJ site."""
        return link_text_obj(self.external_url, self.url_slug)

    def external_links_txt(self, _style: str = '', include_alt_links: bool = True) -> Text:
        """Overrides super() method to apply self.border_style."""
        return super().external_links_txt(self.border_style, include_alt_links=include_alt_links)

    def image_with_no_text_msg(self) -> RenderableType:
        """One line of linked text to show if this file doesn't seem to have any OCR text."""
        return Padding(
            Text('').append(self.doj_link()).append(f" is a single image with no text..."),
            (0, 0, 0, 1)
        )

    def printable_document(self) -> Self | Email:
        """Return a copy of this `DojFile` with simplified text if file ID is in `REPLACEMENT_TEXT`."""
        if Document.is_email(self):
            try:
                return Email(self.file_path, text=self.text)  # Pass text= to avoid reprocessing
            except Exception as e:
                self.warn(f"Error creating Email object, trying full reload of text...")
                return Email(self.file_path)
        else:
            return self

    def strip_image_ocr_panels(self) -> None:
        """Removes the ╭--- Page 5, Image 1 ---- panels from the text."""
        new_text, num_replaced = IMAGE_PANEL_REGEX.subn('', self.text)
        self.warn(f"Stripped {num_replaced} image panels.")
        self._set_computed_fields(text=new_text)

    def _repair(self) -> None:
        """Overloads superclass method."""
        new_text = self.repair_ocr_text(OCR_REPAIRS, self.text)
        self._set_computed_fields(text=new_text)
        self._remove_number_only_lines()

    def _remove_number_only_lines(self) -> None:
        """Remove number only lines (which happen a lot in legal doc OCR) if there are more than a certain amount of them."""
        non_number_lines = [line for line in self.lines if not IGNORE_LINE_REGEX.match(line)]
        number_only_line_count = len(self.lines) - len(non_number_lines)

        if number_only_line_count > 20:
            self.warn(f"Reduced line count from {len(self.lines)} to {len(non_number_lines)}")
            self._set_computed_fields(lines=non_number_lines)

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        doc = self.printable_document()

        # Emails handle their own formatting
        if isinstance(doc, Email):
            yield doc
        else:
            yield (info_panel := self.file_info_panel())
            border_style = info_panel.renderables[0].border_style
            panel_args = [self.prettified_text, border_style]

            if self.panel_title_timestamp:
                panel_args.append(Text(f"[{self.panel_title_timestamp}]", style='dim italic ' + border_style))

            table = LeftBarPanel.build(*panel_args)
            yield Padding(table, (0, 0, 1, 1))



import json
import re
from collections import defaultdict
from dataclasses import dataclass, field

from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.text import Text

from epstein_files.documents.emails.emailers import EMAILER_ID_REGEXES
from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import ARCHIVE_LINK_COLOR
from epstein_files.util.constants import EPSTEIN_V_ROTHSTEIN_EDWARDS, OSBORNE_LLP, REPLY_REGEX, SENT_FROM_REGEX
from epstein_files.util.data import sort_dict, without_falsey
from epstein_files.util.doc_cfg import *
from epstein_files.util.env import args
from epstein_files.util.logging import logger

CIVIL_ATTORNEY = 'civil attorney'
CRIMINAL_DEFENSE_ATTORNEY = 'criminal defense attorney'
CRIMINAL_DEFENSE_2008 = f"{CRIMINAL_DEFENSE_ATTORNEY} on 2008 case"
EPSTEIN_LAWYER = 'lawyer'
EPSTEIN_V_ROTHSTEIN_EDWARDS_ATTORNEY = f"{CIVIL_ATTORNEY} {EPSTEIN_V_ROTHSTEIN_EDWARDS}"
ESTATE_EXECUTOR = 'estate executor'
EPSTEIN_ESTATE_EXECUTOR = f"Epstein {ESTATE_EXECUTOR}"
MC2_MODEL_MANAGEMENT = f"{JEAN_LUC_BRUNEL}'s MC2 Model Management"
MIDEAST = 'mideast'
QUESTION_MARKS_TXT = Text(QUESTION_MARKS, style='grey50')
REGEX_STYLE_PREFIX = 'regex'
SIMPLE_NAME_REGEX = re.compile(r"^[-\w, ]+$", re.IGNORECASE)
TECH_BRO = 'tech bro'

VICTIM_COLOR = 'orchid1'

CATEGORY_STYLE_MAPPING = {
    ARTICLE: JOURNALIST,
    BOOK: JOURNALIST,
    LEGAL: EPSTEIN_LAWYER,
    POLITICS: LOBBYIST,
    PROPERTY: BUSINESS,
    REPUTATION: PUBLICIST,
}

CATEGORY_STYLES = {
    JSON: 'dark_red',
    'letter': 'medium_orchid1'
}

debug_console = Console(color_system='256')


@dataclass(kw_only=True)
class BaseHighlight:
    """
    Regex and style information for things we want to highlight.

    Attributes:
        label (str): RegexHighlighter match group name
        pattern (str): regex pattern identifying strings matching this group
        style (str): Rich style to apply to text matching this group
        theme_style_name (str): The style name that must be a part of the rich.Console's theme
    """
    label: str = ''
    regex: re.Pattern = field(init=False)
    style: str
    theme_style_name: str = field(init=False)
    _capture_group_label: str = field(init=False)
    _match_group_var: str = field(init=False)

    def __post_init__(self):
        if not self.label:
            raise ValueError(f'Missing label for {self}')

        self._capture_group_label = self.label.lower().replace(' ', '_').replace('-', '_')
        self._match_group_var = fr"?P<{self._capture_group_label}>"
        self.theme_style_name = f"{REGEX_STYLE_PREFIX}.{self._capture_group_label}"


@dataclass(kw_only=True)
class HighlightedText(BaseHighlight):
    """
    Color highlighting for things other than people's names (e.g. phone numbers, email headers).

    Attributes:
        label (str): RegexHighlighter match group name, defaults to 1st 'emailers' key if only 1 emailer provided
        patterns (list[str]): regex patterns identifying strings matching this group
    """
    patterns: list[str] = field(default_factory=list)
    _pattern: str = field(init=False)

    def __post_init__(self):
        super().__post_init__()

        if not self.label:
            raise ValueError(f"No label provided for {repr(self)}")

        self._pattern = '|'.join(self.patterns)
        self.regex = re.compile(fr"({self._match_group_var}{self._pattern})", re.IGNORECASE | re.MULTILINE)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(label='{self.label}', pattern='{self._pattern}', style='{self.style}')"


@dataclass(kw_only=True)
class HighlightedNames(HighlightedText):
    """
    Encapsulates info about people, places, and other strings we want to highlight with RegexHighlighter.
    Constructor must be called with either an 'emailers' arg or a 'pattern' arg (or both).

    Attributes:
        category (str): optional string to use as an override for self.label in some contexts
        emailers (dict[str, str | None]): optional names to construct regexes for (values are descriptions)
        _pattern (str): regex pattern combining 'pattern' with first & last names of all 'emailers'
    """
    category: str = ''
    emailers: dict[str, str | None] = field(default_factory=dict)
    should_match_first_last_name: bool = True

    def __post_init__(self):
        if not (self.emailers or self.patterns):
            raise ValueError(f"Must provide either 'emailers' or 'pattern' arg.")
        elif not self.label:
            if len(self.emailers) == 1:
                self.label = [k for k in self.emailers.keys()][0]
            else:
                raise ValueError(f"No label provided for {repr(self)}")

        super().__post_init__()
        self._pattern = '|'.join([self._emailer_pattern(e) for e in self.emailers] + self.patterns)
        self.regex = re.compile(fr"\b({self._match_group_var}({self._pattern})s?)\b", re.IGNORECASE)

    def category_str(self) -> str:
        if self.category:
            return self.category
        elif len(self.emailers) == 1 and self.label == [k for k in self.emailers.keys()][0]:
            return ''
        else:
            return self.label.replace('_', ' ')

    def info_for(self, name: str, include_category: bool = False) -> str | None:
        """Label and additional info for 'name' if 'name' is in self.emailers."""
        info_pieces = [self.category_str()] if include_category else []
        info_pieces.append(self.emailers.get(name) or '')
        info_pieces = without_falsey(info_pieces)
        return ', '.join(info_pieces) if info_pieces else None

    def _emailer_pattern(self, name: str) -> str:
        """Pattern matching 'name'. Extends value in EMAILER_ID_REGEXES with first/last name if it exists."""
        if not self.should_match_first_last_name:
            return name

        if name in EMAILER_ID_REGEXES:
            name_patterns = [EMAILER_ID_REGEXES[name].pattern]
        else:
            name_patterns = [remove_question_marks(name).replace(' ', r"\s+")]

        if ' ' in name:
            for partial_name in [reversed_name(name), extract_first_name(name), extract_last_name(name)]:  # Order matters
                if partial_name.lower() not in NAMES_TO_NOT_HIGHLIGHT and SIMPLE_NAME_REGEX.match(partial_name):
                    name_patterns.append(partial_name.replace(' ', r"\s+"))

        pattern = '|'.join(name_patterns)

        if args.deep_debug and args.colors_only:
            debug_console.print(Text('').append(f"{name:25s}", style=self.style).append(f" '{pattern}'", style='dim'))

        return pattern

    def __str__(self) -> str:
        return super().__str__()

    def __repr__(self) -> str:
        s = f"{type(self).__name__}("

        for property in ['label', 'style', 'category', 'patterns', 'emailers']:
            value = getattr(self, property)

            if not value or (property == 'label' and len(self.emailers) == 1 and not self.patterns):
                continue

            s += f"\n    {property}="

            if isinstance(value, dict):
                s += '{'

                for k, v in value.items():
                    s += f"\n        {constantize_name(k)}: {json.dumps(v).replace('null', 'None')},"

                s += '\n    },'
            elif property == 'patterns':
                s += '[\n        '
                s += repr(value).removeprefix('[').removesuffix(']').replace(', ', ',\n        ')
                s += ',\n    ],'
            else:
                s += f"{json.dumps(value)},"

        return s + '\n)'


@dataclass(kw_only=True)
class ManualHighlight(BaseHighlight):
    """For when you can't construct the regex."""
    pattern: str

    def __post_init__(self):
        super().__post_init__()

        if self._match_group_var not in self.pattern:
            raise ValueError(f"Label '{self.label}' must appear in regex pattern '{self.pattern}'")

        self.regex = re.compile(self.pattern, re.MULTILINE)


HIGHLIGHTED_NAMES = [
    # This has to come first to get both stylings applied to the email subjects
    ManualHighlight(
        label='email_subject',
        style='light_yellow3',
        pattern=r"^(> )?(Classification|Flag|Subject|Sujet ?): (?P<email_subject>.*)",
    ),
    HighlightedNames(
        label=ACADEMIA,
        style='light_goldenrod2',
        emailers={
            'Daniel Kahneman': 'Nobel economic sciences laureate and cognitivie psychologist (?)',
            DAVID_HAIG: 'evolutionary geneticist?',
            'David Grosof': 'MIT Sloan School of Management',
            'Ed Boyden': f'{MIT_MEDIA_LAB} neurobiology professor',
            'Harry Fisch': "men's health expert at New York-Presbyterian / Weill Cornell (?)",
            JOSCHA_BACH: 'cognitive science / AI research',
            LAWRENCE_KRAUSS: 'theoretical physicist with #MeToo problems',
            LINDA_STONE: f'ex-Microsoft, {MIT_MEDIA_LAB}',
            MARK_TRAMO: 'professor of neurology at UCLA',
            'Nancy Dahl': f'wife of {LAWRENCE_KRAUSS}',
            NEAL_KASSELL: 'professor of neurosurgery at University of Virginia',
            NOAM_CHOMSKY: f"professor of linguistics at MIT",
            'Norman Finkelstein': 'scholar, well known critic of Israel',
            PETER_ATTIA: 'longevity medicine',
            ROBERT_TRIVERS: 'evolutionary biology',
            ROGER_SCHANK: 'Teachers College, Columbia University',
            'Valeria Chomsky': f"wife of {NOAM_CHOMSKY}",
        },
        patterns=[
            r"Andy\s*Lippman",  # Media Lab
            r"Arizona\s*State\s*University",
            r"Bard\s+((Early )?College|High School|Schools)",
            r"Brotherton",
            r"Carl\s*Sagan",
            r"Columbia(\s*(Business\s*School|University))?",
            r"Dan(iel|ny) Kahneman",
            r"(Francis\s*)?Crick",
            r"J(ames|im)\s*Watson",
            r"(Lord\s*)?Martin\s*Rees",
            r"Massachusetts\s*Institute\s*of\s*Technology",
            r"Mayo\s*Clinic",
            r"Media\s*Lab",
            r"(Marvin\s*)?Minsky",
            r"MIT(\s*Media\s*Lab)?",
            r"Norman\s*Finkelstein",
            r"Oxford(?! Analytica)",
            r"Praluent",
            r"Princeton(\s*University)?",
            r"Regeneron",
            r"(Richard\s*)?Dawkins",
            r"Rockefeller\s*University",
            r"(Sandy\s*)?Pentland",  # Media Lab
            r"Sanofi",
            r"Stanford(\s*University)?(\s*Hospital)?",
            r"(Ste(ph|v)en\s*)?Hawking",
            r"(Steven?\s*)?Pinker",
            r"Texas\s*A&M",
            r"Tulane",
            r"UCLA",
        ],
    ),
    HighlightedNames(
        label='Africa',
        style='light_pink4',
        emailers={
            'Abdoulaye Wade': "former president of Senegal, father of Karim Wade",
            'Ivan Glasenberg': "South African former CEO of Glencore, one of the world's largest commodity trading and mining companies",
            'Karim Wade': 'son of the president of Senegal, facing arrest for corruption, email handle is "Afri zp"',
            'Miles Alexander': 'Operations Manager Michaelhouse Balgowan KwaZulu-Natal South Africa',
            'Macky Sall': 'prime minister of Senegal, defeated Abdoulaye Wade',
        },
        patterns=[
            r"Buhari",
            r"Econet(\s*Wireless)",
            r"Ethiopian?",
            r"Ghana(ian)?",
            r"Glencore",
            r"Goodluck Jonathan",
            r"Johannesburg",
            r"Kenyan?",
            r"Nigerian?",
            r"Okey Enelamah",
            r"(Paul\s*)?Kagame",
            r"Rwandan?",
            r"Senegal(ese)?",
            r"Serengeti",
            r"(South\s*)?African?",
            r"(Strive\s*)?Masiyiwa",
            r"Tanzanian?",
            r"Ugandan?",
            r"(Yoweri\s*)?Museveni",
            r"Zimbabwe(an)?",
        ],
    ),
    HighlightedNames(
        label=ARTS,
        style='light_steel_blue3',
        emailers={
            ANDRES_SERRANO: "'Piss Christ' artist",
            'Barry Josephson': 'American film producer, editor FamilySecurityMatters.org',
            BILL_SIEGEL: 'documentary film producer and director',
            DAVID_BLAINE: 'famous magician',
            'David Brenner': 'American comedian and actor',
            'Richard Merkin': 'painter, illustrator and arts educator',
            STEVEN_PFEIFFER: 'Associate Director at Independent Filmmaker Project (IFP)',
            'Steven Gaydos': 'American screenwriter and journalist',
        },
        patterns=[
            r"(Art )?Spiegelman",
            r"Artspace",
            r"Ayn\s*Rand",
            r"Bobby slayton",
            r"bono\s*mick",
            r"Errol(\s*Morris)?",
            r"Etienne Binant",
            r"(Frank\s)?Gehry",
            r"Harvey\s*Weinstein", r"wientstein", r"Weinstein\s*Co(s?|mpany)",
            r"IFP",
            r"Independent\s*Filmmaker\s*Project",
            r"Jagger",
            r"(Jeffrey\s*)?Katzenberg",
            r"(Johnny\s*)?Depp",
            r"Kid Rock",
            r"(Larry\s*)?Gagosian",
            r"Lena\s*Dunham",
            r"Madonna",
            r"Mark\s*Burnett",
            r"New York Film Festival",
            r"Peter Getzels",
            r"Phaidon",
            r"Ramsey Elkholy",
            r"Regan arts",
            r"shirley maclaine",
            r"Woody( Allen)?",
            r"Zach Braff",
        ],
    ),
    HighlightedNames(
        label=BILL_GATES,
        style='turquoise4',
        category=TECH_BRO,
        emailers={
            BILL_GATES: 'ex-Microsoft, Gates Foundation, bgC3',
            BORIS_NIKOLIC: f'biotech VC partner of {BILL_GATES}, {EPSTEIN_ESTATE_EXECUTOR}',
        },
        patterns=[
            r"BG",
            r"b?g?C3",
            r"(Bill\s*((and|or|&)\s*Melinda\s*)?)?Gates(\s*Foundation)?",
            r"Kofi\s*Rashid",
            r"Melinda(\s*Gates)?",
            r"Microsoft",
            r"MSFT",
        ],
    ),
    HighlightedNames(
        label='bitcoin',
        style='orange1 bold',
        emailers={
            AMIR_TAAKI: f"bitcoin bro, partner of {BROCK_PIERCE} (?)",
            BROCK_PIERCE: 'crypto bro, Bannon business partner, Tether co-founder, arrested in house full of machine guns & "8,000 fotografias de pornopgraphia infantil"',
            JEFFREY_WERNICK: 'former COO of Parler, involved in numerous crypto companies like Bitforex',
            JEREMY_RUBIN: 'developer/researcher',
            JOI_ITO: f"former head of {MIT_MEDIA_LAB} and MIT Digital Currency Initiative",
            ANTHONY_SCARAMUCCI: 'Skybridge Capital, FTX investor',
        },
        patterns=[
            r"Balaji",
            r"Bart\s*Stephens",  # co-founder, Blockchain Capital
            r"Bioptix",  # Now RIOT Blockchain
            r"bitcoin(\s*Foundation)?",
            r"Brian Forde",
            r"Brock(\s*Pierce)?",
            r"coins?(\s*Center)?",
            r"Cory\s*Fields",  # bitcoin dev
            r"cr[iy]?pto(currenc(y|ies))?",
            r"Digital\s*Currenc(ies|y)(\s*Initiative)?",
            r"e-currency",
            r"(Gavin )?Andress?en",  # bitcoin dev
            r"(Howard\s+)?Lutnic?k",
            r"(Jim\s*)Pallotta",  # Media lab advisory board
            r"Libra",
            r"Madars",
            r"Mi(chael|ke)\s*Novogratz",
            r"(Patrick\s*)?Murck",
            r"(Hester\s*)?Peirce",
            r"(?-i:RIOT)",  # (?-i:) makes it case sensitive
            r"Ron\s*Rivest",
            r"block ?chain(\s*capital)?",
            r"(Ross\s*)?Ulbricht",
            r"Silk\s*Road",
            r"SpanCash",
            r"Tether",
            r"virtual\s*currenc(ies|y)",
            r"Wire\s*ca\n?rd",
            r"Wladimir( van der Laan)?",  # bitcoin dev
            r"(zero\s+knowledge\s+|zk)pro(of|tocols?)",
        ],
    ),
    HighlightedNames(
        label=BUSINESS,
        style='spring_green4',
        emailers={
            ALIREZA_ITTIHADIEH: 'CEO Freestream Aircraft Limited',
            BARBRO_C_EHNBOM: 'Swedish pharmaceuticals, SALSS',
            BARRY_J_COHEN: None,
            'David Mitchell': 'Mitchell Holdings, New York real estate developer',
            GERALD_BARTON: "Maryland property developer Landmark Land Company",
            GORDON_GETTY: 'heir to oil tycoon J. Paul Getty',
            'Philip Kafka': 'president of Prince Concepts (and son of Terry Kafka?)',
            ROBERT_LAWRENCE_KUHN: 'investment banker, China expert',
            TERRY_KAFKA: 'CEO of Impact Outdoor (highway billboards)',
            TOM_PRITZKER: 'chairman of The Pritzker Organization and Hyatt Hotels',
        },
        patterns=[
            r"Arthur Klein",
            r"(Barry\s*)?Honig",
            r"((Bill|David)\s*)?Koch(\s*(Bro(s|thers)|Industries))?",
            r"Gruterite",
            r"((John|Patricia)\s*)?Kluge",
            r"Marc Rich",
            r"(Mi(chael|ke)\s*)?Ovitz",
            r"(Steve\s+)?Wynn",
            r"(Les(lie)?\s+)?Wexner",
            r"Michael\s*Klein",
            r"New Leaf Ventures",
            r"Park Partners",
            r"SALSS",
            r"Swedish[-\s]*American\s*Life\s*Science\s*Summit",
            r"Trilateral Commission",
            r"Valhi",
            r"(Yves\s*)?Bouvier",
        ],
    ),
    HighlightedNames(
        label='cannabis',
        style='chartreuse2',
        patterns=[
            r"CBD",
            r"cannabis",
            r"marijuana",
            r"psychedelic",
            r"THC",
            r"WEED(guide|maps)?[^s]?",
        ],
    ),
    HighlightedNames(
        label='China',
        style='bright_red',
        emailers={
            'Gino Yu': 'professor / game designer in Hong Kong',
        },
        patterns=[
            r"Ali.?baba",
            r"Beijing",
            r"CCP",
            r"Chin(a|e?se)(?! Daily)",
            r"DPRK",
            r"Global Times",
            r"Guo",
            r"Hong",
            r"Huaw[ae]i",
            r"Kim\s*Jong\s*Un",
            r"Kong",
            r"Jack\s+Ma",
            r"Kwok",
            r"Ministry\sof\sState\sSecurity",
            r"Mongolian?",
            r"MSS",
            r"North\s*Korean?",
            r"Peking",
            r"PRC",
            r"Pyongyang",
            r"SCMP",
            r"Xi(aomi)?",
            r"Jinping",
        ],
    ),
    HighlightedNames(
        label='deepak',
        style='dark_sea_green4',
        emailers={
            CAROLYN_RANGEL: f"Deepak Chopra's assistant {QUESTION_MARKS}",
            DEEPAK_CHOPRA: 'woo woo',
        },
    ),
    HighlightedNames(
        label='Democrat',
        style='sky_blue1',
        emailers={
            PAUL_PROSPERI: 'friend of Bill Clinton',
        },
        patterns=[
            r"(Al\s*)?Franken",
            r"Al\s*Gore",
            r"(Barac?k )?Obama",
            r"((Bill|Hillart?y)\s*)?Clinton",
            r"((Chuck|Charles)\s*)?S(ch|hc)umer",
            r"Debbie\s*Wasserman\s*Schultz",
            r"Dem(ocrat(ic)?)?",
            r"(Diana\s*)?DeGette",
            r"DNC",
            r"(Ed(ward)?\s*)?Mezvinsky",
            r"Elena\s*Kagan",
            r"(Eliott?\s*)?Spitzer(, Eliot)?",
            r"Eric Holder",
            r"George\s*Mitchell",
            r"(George\s*)?Soros",
            r"Hakeem\s*Jeffries",
            r"Hill?ary",
            r"HRC",
            r"(Jo(e|seph)\s*)?Biden",
            r"(John\s*)?Kerry",
            r"Lisa Monaco",
            r"(Matteo\s*)?Salvini",
            r"Maxine\s*Waters",
            r"(Nancy )?Pelosi",
            r"Open Society( Global Board)?",
            r"Ron\s*Dellums",
            r"Schumer",
            r"(Tim(othy)?\s*)?Geithner",
            r"Tom McMillen",
            r"Vernon\s*Jordan",
        ],
    ),
    HighlightedNames(
        label='Dubins',
        style='medium_orchid1',
        emailers={
            GLENN_DUBIN: "Highbridge Capital Management, married to Epstein's ex-gf Eva",
            EVA: "possibly Epstein's ex-girlfriend (?)",
            'Eva Dubin': f"Epstein's ex-girlfriend now married to {GLENN_DUBIN}",
        },
        patterns=[r"((Celina|Eva( Anderss?on)?|Glenn?) )?Dubin"],
    ),
    HighlightedNames(
        label='employee',
        style='medium_purple4',
        emailers={
            'Alfredo Rodriguez': "Epstein's butler, stole the journal",
            'Bernard Kruger': "Epstein's doctor",
            EDUARDO_ROBLES: f'home builder at Creative Kingdom Dubai',
            ERIC_ROTH: 'jet decorator at International Jet',
            GWENDOLYN_BECK: 'Epstein fund manager in the 90s',
            JANUSZ_BANASIAK: "Epstein's house manager",
            "John Allessi": "Epstein's houseman",
            JEAN_HUGUEN: 'interior design at Alberto Pinto Cabinet',
            LAWRANCE_VISOSKI: "Epstein's pilot",
            LESLEY_GROFF: f"Epstein's assistant",
            'Linda Pinto': 'interior design at Alberto Pinto Cabinet',
            MERWIN_DELA_CRUZ: None,  # HOUSE_OVERSIGHT_032652 Groff says "Jojo and Merwin both requested off Nov. 25 and 26"
            NADIA_MARCINKO: "Epstein's pilot",
            'Sean J. Lancaster': 'airplane reseller',
        },
        patterns=[
            r"Adriana\s*Ross",
            r"Merwin",
            r"(Sarah\s*)?Kellen", r"Vickers",  # Married name is Metiers
        ],
    ),
    HighlightedNames(
        label='Epstein',
        style='blue1',
        emailers={
            JEFFREY_EPSTEIN: None,
            MARK_EPSTEIN: 'brother of Jeffrey',
        },
        patterns=[
            r"JEGE(\s*Inc)?",
            r"LSJ",
            r"Zorro(\s*Ranch)?",
        ],
    ),
    HighlightedNames(
        label=EPSTEIN_LAWYER,
        style='purple',
        emailers={
            'Alan S Halperin': 'partner at Paul, Weiss',
            ALAN_DERSHOWITZ: 'Harvard Law School professor',
            ARDA_BESKARDES: 'NYC immigration attorney allegedly involved in sex-trafficking operations',
            BENNET_MOSKOWITZ: f'represented the {EPSTEIN_ESTATE_EXECUTOR}s',
            BRAD_KARP: 'head of the law firm Paul Weiss',
            CHRISTIAN_EVERDELL: f"{GHISLAINE_MAXWELL}'s lawyer ca. 2021, Cohen & Gresser",
            'Connie Zaguirre': f"office of {ROBERT_D_CRITTON_JR}",
            DAVID_SCHOEN: f"{CRIMINAL_DEFENSE_ATTORNEY} after 2019 arrest",
            DEBBIE_FEIN: EPSTEIN_V_ROTHSTEIN_EDWARDS_ATTORNEY,
            'Erika Kellerhals': 'attorney in St. Thomas',
            FRED_HADDAD: "co-founder of Heck's in West Virginia",
            GERALD_LEFCOURT: f'friend of {ALAN_DERSHOWITZ}',
            'Howard Rubenstein': f"Epstein's former spokesman",
            JACK_GOLDBERGER: CRIMINAL_DEFENSE_2008,
            JACKIE_PERCZEK: CRIMINAL_DEFENSE_2008,
            JAY_LEFKOWITZ: f"Kirkland & Ellis partner, {CRIMINAL_DEFENSE_2008}",
            JESSICA_CADWELL: f'paralegal to {ROBERT_D_CRITTON_JR}',  # house_oversight_030464
            KEN_STARR: 'head of the Monica Lewinsky investigation against Bill Clinton',
            LILLY_SANCHEZ: CRIMINAL_DEFENSE_ATTORNEY,
            MARTIN_WEINBERG: CRIMINAL_DEFENSE_ATTORNEY,
            MICHAEL_MILLER: 'Steptoe LLP partner',
            REID_WEINGARTEN: 'Steptoe LLP partner',
            ROBERT_D_CRITTON_JR: CRIMINAL_DEFENSE_ATTORNEY,
            'Robert Gold': 'helped Epstein track down money belonging to Spanish families',
            'Roy Black': CRIMINAL_DEFENSE_2008,
            SCOTT_J_LINK: CRIMINAL_DEFENSE_ATTORNEY,
            STACEY_RICHMAN: f"New York {CRIMINAL_DEFENSE_ATTORNEY}",
            TONJA_HADDAD_COLEMAN: f'{EPSTEIN_V_ROTHSTEIN_EDWARDS_ATTORNEY}',  # relation of Fred Haddad?
        },
        patterns=[
            r"(Barry (E. )?)?Krischer",
            r"dersh",
            r"Kate Kelly",
            r"Kirkland\s*&\s*Ellis",
            r"(Leon\s*)?Jaworski",
            r"Michael J. Pike",
            r"Paul,?\s*Weiss",
            r"Steptoe(\s*& Johnson)?(\s*LLP)?",
            r"Sull(ivan)?\s*(&|and)?\s*Crom(well)?",
            r"Wein(berg|garten)",
        ],
    ),
    HighlightedNames(
        label=ESTATE_EXECUTOR,
        style='purple3 bold',
        category=EPSTEIN_LAWYER,
        emailers={
            DARREN_INDYKE: EPSTEIN_ESTATE_EXECUTOR,
            RICHARD_KAHN: EPSTEIN_ESTATE_EXECUTOR,
        },
    ),
    HighlightedNames(
        label='Europe',
        style='light_sky_blue3',
        emailers={
            ANDRZEJ_DUDA: 'former president of Poland',
            'Caroline Lang': 'daughter of Jack Lang',
            EDWARD_ROD_LARSEN: f"son of {TERJE_ROD_LARSEN}",
            'Fabrice Aidan': f'diplomat who worked with {TERJE_ROD_LARSEN}',
            'Jack Lang': 'former French Minister of National Education',
            MIROSLAV_LAJCAK: 'Russia-friendly Slovakian politician, friend of Steve Bannon',
            PETER_MANDELSON: 'UK politics',
            TERJE_ROD_LARSEN: 'Norwegian diplomat',
            THORBJORN_JAGLAND: 'former prime minister of Norway, Nobel Peace Prize Committee',
        },
        patterns=[
            r"AfD",
            r"Alfa(\s*Bank)",
            r"(Angela )?Merk(el|le)",
            r"Austria",
            r"Belgi(an|um)",
            r"(Benjamin\s*)?Harnwell",
            r"Berlin",
            r"Borge",
            r"Boris\s*Johnson",
            r"Brexit(eers?)?",
            r"Brit(ain|ish)",
            r"Brussels",
            r"Cannes",
            r"Cypr(iot|us)",
            r"David\s*Cameron",
            r"Davos",
            r"ECB",
            r"England",
            r"EU",
            r"Europe(an)?(\s*Union)?",
            r"Fr(ance|ench)",
            r"Geneva",
            r"Germany?",
            r"Gillard",
            r"Gree(ce|k)",
            r"Ibiza",
            r"Ital(ian|y)",
            r"Jacques",
            r"Kiev",
            r"Latvian?",
            r"Lithuanian?",
            r"Le\s*Pen",
            r"London",
            r"Macron",
            r"Melusine",
            r"MI\s*5",
            r"Munich",
            r"NATO",
            r"(Nicholas\s*)?Sarkozy",
            r"Nigel(\s*Farage)?",
            r"(Northern\s*)?Ireland",
            r"Norw(ay|egian)",
            r"Oslo",
            r"Paris",
            r"Polish",
            r"pope",
            r"Portugal",
            r"Scotland",
            r"(Sebastian )?Kurz",
            r"Stockholm",
            r"Strasbourg",
            r"Strauss[- ]?Kahn",
            r"Swed(en|ish)(?![-\s]+American Life Scienc)",
            r"Swi(ss|tzerland)",
            r"(Tony\s)?Blair",
            r"United\s*Kingdom",
            r"U\.K\.",
            r"Ukrain(e|ian)",
            r"Venice",
            r"(Vi(c|k)tor\s+)?Orbah?n",
            r"Vienna",
            r"Zug",
            r"Zurich",
        ],
    ),
    HighlightedNames(
        label=FINANCE,
        style='green',
        emailers={
            AMANDA_ENS: 'Citigroup',
            BRAD_WECHSLER: f"head of {LEON_BLACK}'s personal investment vehicle according to FT",
            CECILIA_STEEN: None,
            DANIEL_SABBA: 'UBS Investment Bank',
            DAVID_FISZEL: 'CIO Honeycomb Asset Management',
            JES_STALEY: 'former CEO of Barclays',
            JIDE_ZEITLIN: 'former partner at Goldman Sachs, allegations of sexual misconduct',
            'Laurie Cameron': 'currency trading',
            LEON_BLACK: 'Apollo CEO who paid Epstein tens of millions for tax advice',
            MARC_LEON: 'Luxury Properties Sari Morrocco',
            MELANIE_SPINELLA: 'representative of Leon Black',
            MORTIMER_ZUCKERMAN: 'business partner of Epstein, newspaper publisher',
            NORMAN_D_RAU: 'managing director at Morgan Stanley',
            PAUL_BARRETT: None,
            PAUL_MORRIS: DEUTSCHE_BANK,
            'Skip Rimer': 'Milken Institute (Michael Milken)',
            'Steven Elkman': DEUTSCHE_BANK,
            'Vahe Stepanian': 'Cetera Financial Group',
            VINIT_SAHNI: f"analyst at {DEUTSCHE_BANK} and {GOLDMAN_SACHS}",
        },
        patterns=[
            r"Ace\s*Greenberg",
            r"AIG",
            r"((anti.?)?money\s+)?launder(s?|ers?|ing)?(\s+money)?",
            r"Apollo",
            r"Ari\s*Glass",
            r"Bank(\s*of\s*Scotland)",
            r"Bear\s*Stearns",
            r"(Bernie\s*)?Madoff",
            r"Black(rock|stone)",
            r"B\s*of\s*A",
            r"Boothbay(\sFund\sManagement)?",
            r"Chase\s*Bank",
            r"Conrad B",
            r"Credit\s*Suisse",
            r"DB",
            r"Deutsche?\s*(Asset|Bank)",
            r"Electron\s*Capital\s*(Partners)?",
            r"Fenner",
            r"FRBNY",
            r"Goldman(\s*Sachs)",
            r"GRAT",
            r"Gratitude (America|& Enhanced)",  # Leon Black and/or Epstein charity?
            r"Hank\s*Greenburg",
            r"HSBC",
            r"Invesco",
            r"Jamie\s*D(imon)?",
            r"(Janet\s*)?Yellen",
            r"(Jerome\s*)?Powell(?! M\. Cabot)",
            r"(Jimmy\s*)?Cayne",
            r"Joon\s*Yun",
            r"JPMC?",
            r"j\.?p\.?\s*morgan(\.?com|\s*Chase)?",
            r"Madoff",
            r"Merrill(\s*Lynch)?",
            r"(Michael\s*)?Cembalest",
            r"(Mi(chael|ke)\s*)?Milken(\s*Conference|Institute)?",
            r"Mizrahi\s*Bank",
            r"MLPF&S",
            r"Morgan Stanley",
            r"(Peter L. )?Scher",
            r"(Ray\s*)?Dalio",
            r"(Richard\s*)?LeFrak",
            r"Rockefeller(?! University)(\s*Foundation)?",
            r"SBNY",
            r"Serageldin",
            r"Signature\s*Bank",
            r"(Ste(phen|ve)\s*)?Schwart?z?man",
            r"Susquehanna",
            r"UBS",
            r"us.gio@jpmorgan.com",
            r"Wall\s*Street(?!\s*Jour)",
        ],
    ),
    HighlightedNames(
        label=FRIEND,
        style='tan',
        emailers={
            DANGENE_AND_JENNIE_ENTERPRISE: 'founders of the members-only CORE club',
            DAVID_STERN: f'emailed Epstein from Moscow, knows chairman of {DEUTSCHE_BANK} (?)',
            JONATHAN_FARKAS: "heir to the Alexander's department store fortune",
            'linkspirit': 'Skype username of someone Epstein communicated with',
            'Peter Thomas Roth': 'student of Epstein at Dalton, skincare company founder',
            STEPHEN_HANSON: None,
            TOM_BARRACK: 'long time friend of Trump',
        },
        patterns=[
            r"Andrew Farkas",
            r"Jonanthan and Kimberly Farkus",
            r"Thomas\s*(J\.?\s*)?Barrack(\s*Jr)?",
        ],
    ),
    HighlightedNames(
        label='government',
        style='color(24) bold',
        emailers={
            ALISON_J_NATHAN: "judge in New York's Southern District",
            ANN_MARIE_VILLAFANA: 'Southern District of Florida (SDFL) U.S. Attorney',
            CHRISTOPHER_DILORIO: 'self described whistleblower',
            DANNY_FROST: 'Director of Communications at Manhattan D.A.',
            'Police Code Enforcement': f"{PALM_BEACH} buildings code enforcement",
        },
        patterns=[
            r"AG",
            r"(Alicia\s*)?Valle",
            r'Alice\s*Fisher|Fisher, Alice',
            r"AML",
            r"(Andrew\s*)?(McCabe|Natsios)",
            r"(Assistant\s+)?State\s*Attorney",
            r"Attorney General",
            r'Barbara\s*Burns',  # AUSA
            r"((Bob|Robert)\s*)?Mueller",
            r"(Byung\s)?Pak",
            r"Case 1:19-cv-03377(-LAP)?",
            r"(CENT|NORTH|SOUTH)COM",
            r"CFTC?",
            r"CIA",
            r"CIS",
            r"CVRA",
            r"DARPA",
            r"Dep(artmen)?t\.?\s*of\s*(the\s*)?(Justice|Treasury)",
            r"DHS",
            r"DOJ",
            r"EDGAR (Filing|Search)",  # SEC database is EDGAR
            r"FBI",
            r"FCPA",
            r"FDIC",
            r"FDLE",
            r"Federal\s*Bureau\s*of\s*Investigation",
            r"FinCEN",
            r"(www\.)?FINRA(\.org)?",
            r"FOIA",
            r"FTC",
            r"(General\s*)?P(a|e)traeus",
            r"Geoff\s*Ling",
            r"Homeland\s*Security",
            r"IRS",
            r"(James\s*)?Comey",
            r"(Jennifer\s*Shasky\s*)?Calvery",
            r"((Judge|Mark)\s*)?(Carney|Filip)",
            r"(Judge\s*)?(Kenneth\s*)?(A\.?\s*)?Marra",
            r"(Justice|Treasury)\s*Dep(t|artment)",
            r"(Kirk )?Blouin",
            r"KYC",
            r"(Lann?a\s*)?Belohlavek",
            r"NIH",
            r"NPA",
            r"NS(A|C)",
            r"OCC",
            r"OFAC",
            r"(Michael\s*)?Reiter",
            r"OGE",
            r"Office\s*of\s*Government\s*Ethics",
            r"PBPD",
            r"police",
            r"(Preet\s*)?Bharara",
            r"SCOTUS",
            r"SD(FL|NY)",
            r"SEC(\.gov)?",
            r"Secret\s*Service",
            r"Securities\s*and\s*Exchange\s*Commission",
            r"Southern\s*District(\s*of\s*(Florida|New\s*York))?",
            r"State\s*Dep(artmen)?t",
            r"Strzok",
            r"Supreme\s*Court",
            r"Treasury\s*(Dep(artmen)?t|Secretary)",
            r"TSA",
            r"U\.?S\.? attorney",
            r"USAID",
            r"US\s*(AF|Army|Air\s*Force)",
            r"Walter\s*Reed(\s*Army\s*Institute\s*of\s*Research)?",
            r"(William\s*J\.?\s*)?Zloch",
            r"WRAIR",
        ],
    ),
    HighlightedNames(
        label=HARVARD,
        style='light_goldenrod3',
        emailers={
            'Donald Rubin': 'Professor of Statistics',
            'Kelly Friendly': 'longtime aide and spokesperson of Larry Summers',
            LARRY_SUMMERS: 'board of Digital Currency Group (DCG), Obama economic advisor',
            'Leah Reis-Dennis': "producer for Lisa New's Poetry in America",
            LISA_NEW: f'professor of poetry, wife of {LARRY_SUMMERS}, AKA "Elisa New"',
            'Lisa Randall': 'theoretical physicist',
            MARTIN_NOWAK: 'professor of mathematics and biology',
            MOSHE_HOFFMAN: 'behavioral and evolutionary economics',
        },
        patterns=[
            r"Cambridge",
            r"(Derek\s*)?Bok",
            r"Elisa(\s*New)?",
            r"Harvard(\s*(Business|Law|University)(\s*School)?)?",
            r"(Jonathan\s*)?Zittrain",
            r"Poetry\s*in\s*America",
            r"(Stephen\s*)?Kosslyn",
        ],
    ),
    HighlightedNames(
        label='India',
        style='bright_green',
        emailers={
            ANIL_AMBANI: 'billionaire chairman of Reliance Group',
        },
        patterns=[
            r"Abraaj",
            r"Anna\s*Hazare",
            r"(Arif\s*)?Naqvi",
            r"(Arvind\s*)?Kejriwal",
            r"Bangalore",
            r"Hardeep( Pur[ei]e)?",
            r"Indian?",
            r"InsightsPod",
            r"Modi",
            r"Mumbai",
            r"(New\s*)?Delhi",
            r"Tranchulas",
        ],
    ),
    HighlightedNames(
        label='Israel',
        style='dodger_blue2',
        emailers={
            EHUD_BARAK: 'former prime minister of Israel, Epstein business partner',
            'Mitchell Bard': 'director of the American-Israeli Cooperative Enterprise (AICE)',
            NILI_PRIELL_BARAK: 'wife of Ehud Barak',
        },
        patterns=[
            r"AIPAC",
            r"Bibi",
            r"(eh|(Ehud|Nili Priell)\s*)?barak",
            r"EB",
            r"Ehud\s*Barack",
            r"Hapoalim",
            r"Israeli?",
            r"Jerusalem",
            r"J\s*Street",
            r"Menachem\s*Begin",
            r"Mossad",
            r"Netanyahu",
            r"(Sheldon\s*)?Adelson",
            r"Tel\s*Aviv",
            r"(The\s*)?Shimon\s*Post",
            r"Yitzhak", r"Rabin",
            r"YIVO",
            r"zionist",
        ],
    ),
    HighlightedNames(
        label='Japan',
        style='color(168)',
        patterns=[
            r"BOJ",
            r"(Bank\s+of\s+)?Japan(ese)?",
            r"jpy?(?! Morgan)",
            r"SG",
            r"Singapore",
            r"Toky[op]",
        ],
    ),
    HighlightedNames(
        label=JOURNALIST,
        style='bright_yellow',
        emailers={
            'Alain Forget': 'author of "How To Get Out Of This World ALIVE"',
            'Alex Yablon': 'New York Magazine fact checker (?)',
            EDWARD_JAY_EPSTEIN: 'no relation, wrote books about spies',
            HENRY_HOLT: f"{MICHAEL_WOLFF}'s book publisher (company not a person)",
            JAMES_HILL: 'ABC News',
            JENNIFER_JACQUET: 'Future Science magazine',
            JOHN_BROCKMAN: 'literary agent and author specializing in scientific literature',
            LANDON_THOMAS: 'New York Times financial reporter',
            MICHAEL_WOLFF: 'Author of "Fire and Fury: Inside the Trump White House"',
            PAUL_KRASSNER: '60s counterculture guy',
            'Peter Aldhous': 'Buzzfeed science reporter',
            "Susan Edelman": 'New York Post reporter',
            'Tim Zagat': 'Zagat restaurant guide CEO',
        },
        patterns=[
            r"ABC(\s*News)?",
            r"Alexandra Wolfe|Wolfe, Alexandra",
            r"AlterNet",
            r"Arianna(\s*Huffington)?",
            r"(Arthur\s*)?Kretchmer",
            r'Associated\s*Press',
            r"Axios",
            r"BBC",
            r"(Bob|Robert)\s*(Costa|Woodward)",
            r"Breitbart",
            r"BuzzFeed(\s*News)?",
            r"C-?Span",
            r"CBS(\s*(4|Corp|News))?",
            r"Charlie\s*Rose",
            r"China\s*Daily",
            r"(C|MS)?NBC(\s*News)?",
            r"CNN(politics?)?",
            r"Con[cs]hita", r"Sarnoff",
            r"Daily Business Review",
            r"(?<!Virgin[-\s]Islands[-\s])(The\s*)?Daily\s*(Beast|Mail|News|Telegraph)",
            r"(David\s*)?(Pecker|Pegg)",
            r"David\s*Brooks",
            r"Ed\s*Krassenstein",
            r"(Emily\s*)?Michot",
            r"Ezra\s*Klein",
            r"Fire\s*and\s*Fury",
            r"Forbes",
            r"Fortune\s*Magazine",
            r"Fox\s*News(\.com)?",
            r"FrontPage Magazine",
            r"FT",
            r"(George\s*)?Stephanopoulus",
            r"Ger(ald|ry)\s*Baker",
            r"Globe\s*and\s*Mail",
            r"Good\s*Morning\s*America",
            r"Graydon(\s*Carter)?",
            r"Hollywood\s*Reporter",
            r"Huff(ington)?(\s*Po(st)?)?",
            r"Ingram, David",
            r"James\s*Hill",
            r"(James\s*)?Patterson",
            r"Jesse Kornbluth",
            r"John\s*Connolly",
            r"Jonathan\s*Karl",
            r"Journal of Criminal Law and Criminology",
            r"Julie\s*(K.?\s*)?Brown", r'jbrown@miamiherald.com',
            r"(Katie\s*)?Couric",
            r"Keith\s*Larsen",
            r"L\.?A\.?\s*Times",
            r"Law(360|\.com|fare)",
            r"(Les\s*)?Moonves",
            r"MarketWatch",
            r"Miami\s*Herald",
            r"(Mi(chael|ke)\s*)?Bloomber[gq](\s*News)?",
            r"(Michele\s*)?Dargan",
            r"Morning News USA",
            r"(National\s*)?Enquirer",
            r"News(max|week)",
            r"NYer",
            r"Palm\s*Beach\s*(Daily\s*News|Post)",
            r"PERVERSION\s*OF\s*JUSTICE",
            r"Politico",
            r"Pro\s*Publica",
            r"(Sean\s*)?Hannity",
            r"Sharon Churcher",  # Daily Mail
            r"Sulzberger",
            r"SunSentinel",
            r"(The\s*)?Financial\s*Times",
            r"The\s*Guardian",
            r"TheHill",
            r"(The\s*)?Mail\s*On\s*Sunday",
            r"(The\s*)?N(ew\s*)?Y(ork)?\s*(Magazine|Observer|P(ost)?|T(imes)?)",
            r"(The\s*)?New\s*Yorker",
            r"(The\s*)?Wall\s*Street\s*Journal",
            r"(The\s*)?Wa(shington\s*)?Po(st)?",
            r"(Thomson\s*)?Reuters",
            r"(Uma\s*)?Sanghvi",
            r"USA\s*Today",
            r"Vanity\s*Fair",
            r"Viceland",
            r"Vick[iy]\s*Ward",
            r"Vox",
            r"WGBH",
            r"WSJ",
            r"[-\w.]+@(bbc|independent|mailonline|mirror|thetimes)\.co\.uk",
        ],
    ),
    HighlightedNames(
        label=JUNK,
        style='gray46',
        emailers={
            'asmallworld@travel.asmallworld.net': None,
            "digest-noreply@quora.com": None,
            'editorialstaff@flipboard.com': None,
            'How To Academy': None,
            'Jokeland': None,
        },
        should_match_first_last_name=False,
    ),
    HighlightedNames(
        label='Latin America',
        style='yellow',
        patterns=[
            r"Argentin(a|ian)",
            r"Bolsonar[aio]",
            r"Bra[sz]il(ian)?",
            r"Caracas",
            r"Castro",
            r"Chile",
            r"Colombian?",
            r"Cuban?",
            r"el chapo",
            r"El\s*Salvador",
            r"((Enrique )?Pena )?Nieto",
            r"Lat(in)?\s*Am(erican?)?",
            r"Lula",
            r"(?<!New )Mexic(an|o)",
            r"(Nicolas\s+)?Maduro",
            r"Panama( Papers)?",
            r"Peru(vian)?",
            r"Venezuelan?",
            r"Zambrano",
        ],
    ),
    HighlightedNames(
        label=LOBBYIST,
        style='light_coral',
        emailers={
            BOB_CROWE: 'partner at Nelson Mullins',
            'Joshua Cooper Ramo': 'co-CEO of Henry Kissinger Associates',
            KATHERINE_KEATING: 'daughter of former Australian prime minister',
            MOHAMED_WAHEED_HASSAN: 'former president of the Maldives',
            OLIVIER_COLOM: 'France',
            'Paul Keating': 'former prime minister of Australia',
            PUREVSUREN_LUNDEG: 'Mongolian ambassador to the UN',
            'Stanley Rosenberg': 'former President of the Massachusetts Senate',
        },
        patterns=[
            r"CSIS",
            r"elisabeth\s*feliho",
            r"(Kevin\s*)?Rudd",
            r"Stanley Rosenberg",
            r"Vinoda\s*Basnayake",
        ],
    ),
    HighlightedNames(
        label='locations',
        style='cornsilk1',
        patterns=[
            r"Alabama",
            r"Arizona(?! State University)",
            r"Aspen",
            r"Berkeley",
            r"Boston",
            r"Brooklyn",
            r"California",
            r"Canada",
            r"Cape Cod",
            r"Charlottesville",
            r"Colorado",
            r"Connecticut",
            r"Florida",
            r"Jersey\s*City",
            r"Los Angeles",
            r"Loudoun\s*County?",
            r"Martha's\s*Vineyard",
            r"Miami(?!\s?Herald)",
            r"Nantucket",
            r"New\s*(Jersey|Mexico)",
            r"(North|South)\s*Carolina",
            r"NY(C|\s*State)",
            r"Orange\s*County",
            r"Oregon",
            r"Palo Alto",
            r"Pennsylvania",
            r"Phoenix",
            r"Portland",
            r"San Francisco",
            r"Sant[ae]\s*Fe",
            r"Telluride",
            r"Teterboro",
            r"Texas(?! A&M)",
            r"Toronto",
            r"Tu(sc|cs)on",
            r"Vermont",
            r"Washington(\s*D\.?C)?(?!\s*Post)",
            r"Westchester",
        ],
    ),
    HighlightedNames(
        label=MIDEAST,
        style='dark_sea_green4',
        emailers={
            ANAS_ALRASHEED: 'former information minister of Kuwait (???)',
            AZIZA_ALAHMADI: 'Abu Dhabi Department of Culture & Tourism, assistant of Al Sabbagh',
            RAAFAT_ALSABBAGH: 'Saudi royal advisor',
            SHAHER_ABDULHAK_BESHER: 'Yemeni billionaire',
        },
        patterns=[
            r"Abdulmalik Al-Makhlafi",
            r"Abdullah",
            r"Abu\s+Dhabi",
            r"Afghanistan",
            r"Al[-\s]?Qa[ei]da",
            r"Ahmadinejad",
            r"(Rakhat )?Aliyev",
            r"Arab",
            r"Aramco",
            r"Armenia",
            r"Assad",
            r"Ayatollah",
            r"Bahrain",
            r"Basiji?",
            r"Beirut",
            r"Benghazi",
            r"Brunei",
            r"Cairo",
            r"Chagoury",
            r"Damascus",
            r"Dj[iu]bo?uti",
            r"Doha",
            r"[DB]ubai",
            r"Egypt(ian)?",
            r"Emir(at(es?|i))?",
            r"Erdogan",
            r"Fashi",
            r"Gaddafi",
            r"Gulf\s*Cooperation\s*Council",
            r"GCC",
            r"(Hamid\s*)?Karzai",
            r"Hamad\s*bin\s*Jassim",
            r"Hamas",
            r"Hezbollah",
            r"HBJ",
            r"Hourani",
            r"Houthi",
            r"Imran\s+Khan",
            r"Iran(ian)?([-\s]Contra)?",
            r"Isi[ls]",
            r"Islam(abad|ic|ist)?",
            r"Istanbul",
            r"Kabul",
            r"(Kairat\s*)?Kelimbetov",
            r"kasshohgi",
            r"Kaz(akh|ich)stan",
            r"Kazakh?",
            r"Kh[ao]menei",
            r"Khalid\s*Sheikh\s*Mohammed",
            r"Kh?ashoggi",
            r"KSA",
            r"Leban(ese|on)",
            r"Libyan?",
            r"Mahmoud",
            r"Marra[hk]e[cs]h",
            r"MB(N|S|Z)",
            r"Mid(dle)?\s*East(ern)?",
            r"Mohammed\s+bin\s+Salman",
            r"Morocc(an|o)",
            r"Mubarak",
            r"Muslim(\s*Brotherhood)?",
            r"Nayaf",
            r"Nazarbayev",
            r"Pakistani?",
            r"Omar",
            r"(Osama\s*)?Bin\s*Laden",
            r"Osama(?! al)",
            r"Palestin(e|ian)",
            r"Persian?(\s*Gulf)?",
            r"Riya(dh|nd)",
            r"Saddam",
            r"Salman",
            r"Saudi(\s+Arabian?)?",
            r"Shariah?",
            r"SHC",
            r"sheikh",
            r"shia",
            r"(Sultan\s*)?Yacoub",
            r"Sultan(?! Ahmed)",
            r"Syrian?",
            r"(Tarek\s*)?El\s*Sayed",
            r"Tehran",
            r"Timur\s*Kulibayev",
            r"Tripoli",
            r"Tunisian?",
            r"Turk(ey|ish)?(?!s & Caicos)",
            r"UAE",
            r"((Iraq|Iran|Kuwait|Qatar|Yemen)i?)",
        ],
    ),
    HighlightedNames(
        label='modeling',
        style='pale_violet_red1',
        emailers={
            'Abi Schwinck': f'{MC2_MODEL_MANAGEMENT} {QUESTION_MARKS}',
            DANIEL_SIAD: None,
            FAITH_KATES: 'Next Models co-founder',
            'Gianni Serazzi': 'fashion consultant?',
            HEATHER_MANN: 'South African former model, ex-girlfriend of Prince Andrew (?)',
            JEAN_LUC_BRUNEL: f'MC2 Model Management founder, died by suicide in French jail',
            JEFF_FULLER: f'president of {MC2_MODEL_MANAGEMENT} USA',
            'lorraine@mc2mm.com': f'{MC2_MODEL_MANAGEMENT}',
            'pink@mc2mm.com': f'{MC2_MODEL_MANAGEMENT}',
            MANUELA_MARTINEZ: 'Mega Partners (Brazilian agency)',
            MARIANA_IDZKOWSKA: None,
            'Michael Sanka': f'{MC2_MODEL_MANAGEMENT} {QUESTION_MARKS}',
            'Vladimir Yudashkin': 'director of the 1 Mother Agency',
        },
        patterns=[
            r"\w+@mc2mm.com",
            r"MC2",
            r"(Nicole\s*)?Junkerman",  # Also a venture fund manager now
            r"Tigrane",
        ],
    ),
    HighlightedNames(
        label=PUBLICIST,
        style='orange_red1',
        emailers={
            AL_SECKEL: "Isabel Maxwell's husband, Mindshift conference, fell off a cliff",
            'Barnaby Marsh': 'co-founder of philanthropy services company Saint Partners',
            CHRISTINA_GALBRAITH: f"{EPSTEIN_FOUNDATION} Media/PR, worked with {TYLER_SHEARS}",  # Title in 031441
            IAN_OSBORNE: f'{OSBORNE_LLP} reputation repairer hired in 2011',
            MATTHEW_HILTZIK: 'crisis PR at Hiltzik Strategies',
            MICHAEL_SITRICK: 'crisis PR',
            'Owen Blicksilver': 'OBPR, Inc.',
            PEGGY_SIEGAL: 'socialite, movie promoter',
            'R. Couri Hay': None,
            ROSS_GOW: 'Acuity Reputation Management',
            TYLER_SHEARS: f"{REPUTATION_MGMT}, worked on with {CHRISTINA_GALBRAITH}",
        },
        patterns=[
            r"(Matt(hew)? )?Hiltzi[gk]",
            r"Philip\s*Barden",
            r"PR\s*Newswire",
            REPUTATION_MGMT,
            r"Reputation.com",
            r"(Robert L\. )?Dilenschneider",
        ],
    ),
    HighlightedNames(
        label='Republican',
        style='dark_red bold',
        emailers={
            "Juleanna Glover": 'CEO of D.C. public affairs advisory firm Ridgely|Walsh',
            RUDY_GIULIANI: None,
            TULSI_GABBARD: None,
        },
        patterns=[
            r"Alberto\sGonzale[sz]",
            r"(Alex\s*)?Acosta",
            r"(Ben\s*)?Sasse",
            r"Betsy Devos",
            r"((Bill|William)\s*)?Barr",
            r"Bill\s*Shine",
            r"Blackwater",
            r"(Bob\s*)?Corker",
            r"(Brett\s*)?Kavanaugh",
            r"Broidy",
            r"(Chris\s)?Christie",
            r"(?<!Merwin Dela )Cruz",
            r"Darrell\s*Issa",
            r"Devin\s*Nunes",
            r"(Don\s*)?McGa[hn]n",
            r"Erik Prince",
            r"Gary\s*Cohn",
            r"George\s*(H\.?\s*)?(W\.?\s*)?Bush",
            r"(George\s*)?Nader",
            r"GOP",
            r"((Chair|Jay|Joseph)\s*)?Clayton",  # SEC chair, now SDNY
            r"((Bill|William)\s*)?Hinman"
            r"Jeff(rey)?\s*Sessions",
            r"(John\s*(R.?\s*)?)?Bolton",
            r"Kasich",
            r"Keith\s*Schiller",
            r"Kelly(\s*Anne?)?\s*Conway|Kellyanne",
            r"Kissinger",
            r"Kobach",
            r"Kolfage",
            r"(Larry\s*)?Kudlow",
            r"Lewandowski",
            r"(Marco\s)?Rubio",
            r"(Mark\s*)Meadows",
            r"Mattis",
            r"McCain",
            r"McMaster",
            r"(Michael\s)?Hayden",
            r"((General|Mike)\s*)?(Flynn|Pence)",
            r"(Mitt\s*)?Romney",
            r"(Steven?\s*)?Mnuchin",
            r"(Newt\s*)Gingrich",
            r"Nikki",
            r"Haley",
            r"(Paul\s*)?(Manafort|Volcker)",
            r"(Peter\s*)?Navarro",
            r"Pompeo",
            r"Reagan",
            r"Reince", r"Priebus",
            r"Republican",
            r"(Rex\s*)?Till?erson",
            r"(?<!Cynthia )(Richard\s*)?Nixon",
            r"RNC",
            r"(Roy|Stephen)\s*Moore",
            r"Tea\s*Party",
            r"Wilbur\s*Ross",
        ],
    ),
    HighlightedNames(
        label='Rothschild',
        style='indian_red',
        emailers={
            ARIANE_DE_ROTHSCHILD: 'heiress',
            JOHNNY_EL_HACHEM: f'Edmond de Rothschild Private Equity',
        },
        patterns=['AdR'],
    ),
    HighlightedNames(
        label='Russia',
        style='red bold',
        emailers={
            'Dasha Zhukova': 'art collector, daughter of Alexander Zhukov',
            MASHA_DROKOVA: 'silicon valley VC, former Putin Youth member',
            RENATA_BOLOTOVA: 'former model, fund manager at New York State Insurance Fund, Рената Болотова',
            SVETLANA_POZHIDAEVA: "Epstein's Russian assistant who was recommended for a visa by Sergei Belyakov (FSB) and David Blaine",
        },
        patterns=[
            r"Alfa\s*Bank",
            r"Anya\s*Rasulova",
            r"Chernobyl",
            r"Crimea",
            r"Day\s+One\s+Ventures",
            r"(Dmitry\s)?(Kiselyov|(Lana\s*)?Pozhidaeva|Medvedev|Rybolo(o?l?ev|vlev))",
            r"Dmitry",
            r"FSB",
            r"GRU",
            r"KGB",
            r"Kislyak",
            r"Kremlin",
            r"(Anastasia\s*)?Kuznetsova",
            r"Lavrov",
            r"Lukoil",
            r'(Semion\s*)?Mogilevich',
            r"Moscow",
            r"(Natalia\s*)?Veselnitskaya",
            r"(Oleg\s*)?Deripaska",
            r"Oleksandr Vilkul",
            r"Onexim",  # Prokhorov investment vehicle
            r"Prokhorov",
            r"Rakishev",
            r"Rosneft",
            r"RT",
            r"St.?\s*?Petersburg",
            r'Svet',
            r"Russ?ian?",
            r"Sberbank",
            r"Soviet(\s*Union)?",
            r"USSR",
            r"Vlad(imir)?(?! Yudash)",
            r"(Vladimir\s*)?Putin",
            r"Women\s*Empowerment",
            r"Xitrans",
            r"(Vitaly\s*)?Churkin",
        ],
    ),
    HighlightedNames(
        label='Southeast Asia',
        style='light_salmon3 bold',
        patterns=[
            r"Australian?(?! Ave)",
            r"Bangkok",
            r"Burm(a|ese)",
            r"Cambodian?",
            r"Laos",
            r"Malaysian?",
            r"Maldives",
            r"Myan?mar",
            r"New\s*Zealand",
            r"Philippines",
            r"South\s*Korean?",
            r"Tai(pei|wan)",
            r"Thai(land)?",
            r"Vietnam(ese)?",
        ],
    ),
    HighlightedNames(
        label=TECH_BRO,
        style='bright_cyan',
        emailers={
            'Auren Hoffman': 'CEO of SafeGraph (firm that gathers location data from mobile devices) and LiveRamp',
            ELON_MUSK: 'father of Mecha-Hitler',
            PETER_THIEL: 'Paypal mafia member, founder of Palantir, Facebook investor',
            REID_HOFFMAN: 'PayPal mafia member, founder of LinkedIn',
            STEVEN_SINOFSKY: 'ex-Microsoft, loves bitcoin',
            VINCENZO_IOZZO: 'CEO of the identity-security company SlashID',
            ZUBAIR_KHAN: 'Tranchulas cybersecurity, InsightsPod founder, Islamabad / Dubai',
        },
        patterns=[
            r"AG?I",
            r"Artificial\s*(General\s*)?Intelligence",
            r"Chamath", r"Palihapitiya",
            r"Danny\s*Hillis",
            r"deep learning",
            r"Drew\s*Houston",
            r"Eliezer\s*Yudkowsky",
            r"Eric\s*Schmidt",
            r"Greylock(\s*Partners)?",
            r"(?<!(ustin|Moshe)\s)Hoffmand?",
            r"(Jeff\s*)?Bezos",
            r"Larry Page",
            r"LinkedIn",
            r"(Marc\s*)?Andreess?en",
            r"(Mark\s*)?Zuckerberg",
            r"Masa(yoshi)?(\sSon)?",
            r"Najeev",
            r"Nathan\s*Myhrvold",
            r"Palantir",
            r"(Peter\s)?Th(ie|ei)l",
            r"Pierre\s*Omidyar",
            r"Sergey\s*Brin",
            r"Silicon\s*Valley",
            r"Softbank",
            r"SpaceX",
            r"Tim\s*Ferriss?",
            r"Vision\s*Fund",
            r"WikiLeak(ed|s)",
        ],
    ),
    HighlightedNames(
        label='Trump',
        style='red3 bold',
        emailers={
            'Bruce Moskowitz': "'Trump's health guy' according to Epstein",
            NICHOLAS_RIBIS: 'Hilton CEO, former president of Trump Organization',
        },
        patterns=[
            r"@?realDonaldTrump",
            r"(Alan\s*)?Weiss?elberg",
            r"Alex\s*Jones",
            r"(Brad(ley)?\s*)Parscale",
            r"\bDJ?T\b",
            r"Donald J. Tramp",
            r"(Donald\s+(J\.\s+)?)?Trump(ism|\s*(Org(anization)?|Properties)(\s*LLC)?)?",
            r"Don(ald| *Jr)(?! (B|Rubin))",
            r"Ivank?a",
            r"Jared", r"(?<!Tony )Kushner",
            r"(Madeleine\s*)?Westerhout",
            r"Mar[-\s]*a[-\s]*Lago",
            r"(Marla\s*)?Maples",
            r"(Matt(hew)? )?Calamari",
            r"\bMatt C\b",
            r"Michael\s*(D\.?\s*)?Cohen",
            r"Melania",
            r"(Michael (J.? )?)?Boccio",
            r"Paul Rampell",
            r"Rebekah\s*Mercer",
            r"Roger\s+Stone",
            r"rona",
            r"(The\s*)?Art\s*of\s*the\s*Deal",
        ],
    ),
    HighlightedNames(
        label='USVI',
        style='sea_green1',
        emailers={
            CECILE_DE_JONGH: 'Virgin Islands first lady 2007-2015',
            KENNETH_E_MAPP: 'Virgin Islands Governor',
            STACEY_PLASKETT: 'Virgin Islands non-voting member of Congress',
        },
        patterns=[
            r"Antigua",
            r"Bahamas",
            r"BVI",
            r"Caribb?ean",
            r"Dominican\s*Republic",
            r"(Great|Little)\s*St.?\s*James",
            r"Haiti(an)?",
            r"Jamaican?",
            r"(John\s*)deJongh(\s*Jr\.?)",
            r"(Kenneth E\. )?Mapp",
            r"PBI",
            r"Puerto\s*Ric(an|o)",
            r"San\s*Juan",
            r"S(ain)?t.?\s*Thomas",
            r"USVI",
            r"(?<!stein |vis-a-)VI(?!s-a-)",
            r"(The\s*)?Virgin\s*Is(al|la)nd?s(\s*Daily\s*News)?",  # Hard to make this work right
            r"(West\s*)?Palm\s*Beach(\s*County)?(?!\s*(Daily|Post))",
        ],
    ),
    HighlightedNames(
        label='victim',
        style=VICTIM_COLOR,
        patterns=[
            r"child\s*pornography",
            r"(David\s*)?Bo[il]es(,?\s*Schiller( & Flexner)?)?",
            r"(Gloria\s*)?Allred",
            r"(Jane|Tiffany)\s*Doe",
            r"Katie\s*Johnson",
            r"Minor\s*Victim",
            r"pedophile",
            r"Stephanie\s*Clifford",
            r"Stormy\s*Daniels",
            r"(Virginia\s+((L\.?|Roberts)\s+)?)?Giuffre",
            r"Virginia\s+Roberts",
        ],
    ),
    HighlightedNames(
        label='victim lawyer',
        style='medium_orchid1',
        emailers={
            BRAD_EDWARDS: ROTHSTEIN_ROSENFELDT_ADLER,
            'Douglas Wigdor': f'lawsuit against {LEON_BLACK}, Wigdor LLP',
            'Grant J. Smith': ROTHSTEIN_ROSENFELDT_ADLER,
            'Jeanne M. Christensen': f'lawsuit against {LEON_BLACK}, Wigdor LLP',
            JACK_SCAROLA: 'Searcy Denney Scarola Barnhart & Shipley',
            KEN_JENNE: ROTHSTEIN_ROSENFELDT_ADLER,
        },
        patterns=[
            r"(Alan(\s*P.)?|MINTZ)\s*FRAADE",
            r"(J\.?\s*)?(Stan(ley)?\s*)?Pottinger",
            r"(Mi(chael|ke)\s*)?Avenatti",
            r"Paul\s*(G.\s*)?Cassell",
            r"Rothstein\s*Rosenfeldt\s*Adler",
            r"(Scott\s*)?Rothstein",
            r"Wigdor(Law)?",
        ],
    ),
    HighlightedNames(
        label=STEVE_BANNON,
        style='color(58)',
        category='Republican',
        emailers={
            SEAN_BANNON: f"{STEVE_BANNON}'s brother",
            STEVE_BANNON: "Trump campaign manager, memecoin grifter",
        },
        patterns=[
            r"(American\s*)?Dharma",
            r"Biosphere",
        ],
    ),

    # Individuals
    HighlightedNames(
        emailers={STEVEN_HOFFENBERG: "Epstein's Towers Financial ponzi partner, prison for 18 years"},
        style='dark_olive_green3',
        category=FINANCE,
        patterns=[r"(steven?\s*)?hoffenberg?w?"],
    ),
    HighlightedNames(
        emailers={GHISLAINE_MAXWELL: "Epstein's girlfriend, daughter of the spy Robert Maxwell"},
        category='Epstein',
        patterns=[
            r"gmax(1@ellmax.com)?",
            r"(The )?TerraMar Project",
            r"(Scott\s*)?Borgenson",
        ],
        style='deep_pink3',
    ),
    HighlightedNames(emailers={JABOR_Y: '"an influential man in Qatar"'}, category=MIDEAST, style='spring_green1'),
    HighlightedNames(emailers={KATHRYN_RUEMMLER: 'former Obama legal counsel'}, style='magenta2', category=FRIEND),
    HighlightedNames(emailers={MELANIE_WALKER: f"doctor, friend of {BILL_GATES}"}, style='pale_violet_red1', category=FRIEND),
    HighlightedNames(emailers={PAULA: "Epstein's ex-girlfriend who is now in the opera world"}, label='paula', style='pink1', category=FRIEND),
    HighlightedNames(emailers={PRINCE_ANDREW: 'British royal family'}, style='dodger_blue1', category='Europe'),
    HighlightedNames(emailers={SOON_YI_PREVIN: 'wife of Woody Allen'}, style='hot_pink', category=ARTS),
    HighlightedNames(emailers={SULTAN_BIN_SULAYEM: 'chairman of ports in Dubai, CEO of DP World'}, style='green1', category=MIDEAST),

    # HighlightedText not HighlightedNames bc of word boundary issue
    HighlightedText(
        label='metoo',
        style=VICTIM_COLOR,
        patterns=[r"#metoo"]
    ),
    HighlightedText(
        label='phone_number',
        style='bright_green',
        patterns=[
            r"\+?(1?\(?\d{3}\)?[- ]\d{3}[- ]\d{4}|\d{2}[- ]\(?0?\)?\d{2}[- ]\d{4}[- ]\d{4})",
            r"(\b|\+)[\d+]{10,12}\b",
        ],
    ),
    HighlightedText(
        label='unknown',
        style='cyan',
        patterns=[r'\(unknown\)']
    ),
]

# Highlight regexes for things other than names, only used by RegexHighlighter pattern matching
HIGHLIGHTED_TEXTS = [
    HighlightedText(
        label='header_field',
        style='plum4',
        patterns=[r'^[>• ]{,4}(Date ?|From|Sent|To|C[cC]|Importance|Reply[- ]?To|Subject|Bee|B[cC]{2}|Attachments|Flag|Classification|[Il]nline-[Il]mages|((A|Debut du message transfer[&e]|De(stinataire)?|Envoye|Expe(cl|d)iteur|Objet|Q|Sujet) ?)):|^on behalf of'],
    ),
    HighlightedText(
        label='http_links',
        style=f'{ARCHIVE_LINK_COLOR} underline',
        patterns=[r"https?:[^\s]+"],
    ),
    HighlightedText(
        label='quoted_reply_line',
        style='dim',
        patterns=[
            REPLY_REGEX.pattern, r"^(> )?wrote:$",
            r"CONFIDENTIAL FOR ATTORNEY'S EYES ONLY(\nDO NOT COPY)?",
            r"PRIVILEGED ?- ?ATTORNEY WORK.*(\nCONFIDENTIAL - SUBJECT TO.*)?",
        ],
    ),
    HighlightedText(
        label='redacted',
        style='grey58',
        patterns=[fr"{REDACTED}|<?Privileged - Redacted>?"],
    ),
    HighlightedText(
        label='sent_from',
        style='light_cyan3 italic dim',
        patterns=[SENT_FROM_REGEX.pattern],
    ),
    HighlightedText(
        label='snipped_signature',
        style='gray19',
        patterns=[r'<\.\.\.(snipped|trimmed).*\.\.\.>'],
    ),
    HighlightedText(
        label='timestamp_2',
        style=TIMESTAMP_STYLE,
        patterns=[r"\d{1,4}[-/]\d{1,2}[-/]\d{2,4} \d{1,2}:\d{2}:\d{2}( [AP]M)?"],
    ),

    # Manual regexes
    ManualHighlight(
        label='email_attachments',
        style='gray30 italic',
        pattern=r"^(> )?(Attachments|[Il]nline-[Il]mages): (?P<email_attachments>.*)",
    ),
    ManualHighlight(
        label='email_timestamp',
        style=TIMESTAMP_STYLE,
        pattern=r"^(> )?(Date|Sent): (?P<email_timestamp>.*)",
    ),
]

ALL_HIGHLIGHTS = HIGHLIGHTED_NAMES + HIGHLIGHTED_TEXTS
JUNK_EMAILERS = [name for name in [hn for hn in HIGHLIGHTED_NAMES if hn.label == JUNK][0].emailers.keys()]


class EpsteinHighlighter(RegexHighlighter):
    """Finds and colors interesting keywords based on the above config."""
    base_style = f"{REGEX_STYLE_PREFIX}."
    highlights = [highlight_group.regex for highlight_group in ALL_HIGHLIGHTS]
    highlight_counts = defaultdict(int)

    def highlight(self, text: Text) -> None:
        """overrides https://rich.readthedocs.io/en/latest/_modules/rich/highlighter.html#RegexHighlighter"""
        highlight_regex = text.highlight_regex

        for re_highlight in self.highlights:
            highlight_regex(re_highlight, style_prefix=self.base_style)

            if args.debug and isinstance(re_highlight, re.Pattern):
                for match in re_highlight.finditer(text.plain):
                    type(self).highlight_counts[(match.group(1) or 'None').replace('\n', ' ')] += 1

    def print_highlight_counts(self, console: Console) -> None:
        """Print counts of how many times strings were highlighted."""
        highlight_counts = deepcopy(self.highlight_counts)
        weak_date_regex = re.compile(r"^(\d\d?/|20|http|On ).*")

        for highlighted, count in sort_dict(highlight_counts):
            if highlighted is None or weak_date_regex.match(highlighted):
                continue

            try:
                console.print(f"{highlighted:25s} highlighted {count} times")
            except Exception as e:
                logger.error(f"Failed to print highlight count {count} for {highlighted}")


def get_highlight_group_for_name(name: str | None) -> HighlightedNames | None:
    if name is None:
        return None

    for highlight_group in HIGHLIGHTED_NAMES:
        if highlight_group.regex.search(name):
            return highlight_group


def get_style_for_category(category: str) -> str | None:
    if category in CATEGORY_STYLES:
        return CATEGORY_STYLES[category]
    elif category == CONFERENCE:
        return f"{get_style_for_category(ACADEMIA)} dim"
    elif category == SOCIAL:
        return get_style_for_category(PUBLICIST)

    for highlight_group in HIGHLIGHTED_NAMES:
        if highlight_group.label == CATEGORY_STYLE_MAPPING.get(category, category):
            return highlight_group.style


def get_style_for_name(name: str | None, default_style: str = DEFAULT_NAME_STYLE, allow_bold: bool = True) -> str:
    highlight_group = get_highlight_group_for_name(name or UNKNOWN)
    style = highlight_group.style if highlight_group else default_style
    style = style if allow_bold else style.replace('bold', '').strip()
    logger.debug(f"get_style_for_name('{name}', '{default_style}', '{allow_bold}') yielded '{style}'")
    return style


def styled_category(category: str | None) -> Text:
    if not category:
        return QUESTION_MARKS_TXT

    category_str = 'resumé' if category == 'resume' else category.lower()
    return Text(category_str, get_style_for_category(category) or 'wheat4')


def styled_name(name: str | None, default_style: str = DEFAULT_NAME_STYLE) -> Text:
    return Text(name or UNKNOWN, style=get_style_for_name(name, default_style=default_style))


def _print_highlighted_names_repr() -> None:
    for hn in HIGHLIGHTED_NAMES:
        if isinstance(hn, HighlightedNames):
            print(indented(repr(hn)) + ',')
            print(f"pattern: '{hn.regex.pattern}'")

    import sys
    sys.exit()

#_print_highlighted_names_repr()



from dataclasses import dataclass, field
from datetime import datetime, date
from typing import Sequence

from rich.console import Group, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.document import Document
from epstein_files.documents.email import TRUNCATE_EMAILS_FROM, MAILING_LISTS, JUNK_EMAILERS, Email
from epstein_files.documents.messenger_log import MessengerLog
from epstein_files.documents.other_file import OtherFile
from epstein_files.util.constant.strings import *
from epstein_files.util.constant.urls import *
from epstein_files.util.constants import *
from epstein_files.util.data import days_between, flatten, uniquify, without_falsey
from epstein_files.util.env import args
from epstein_files.util.highlighted_group import (QUESTION_MARKS_TXT, HighlightedNames,
     get_highlight_group_for_name, get_style_for_name, styled_category, styled_name)
from epstein_files.util.rich import (GREY_NUMBERS, SKIPPED_FILE_MSG_PADDING, TABLE_TITLE_STYLE, build_table,
     console, join_texts, print_centered)

ALT_INFO_STYLE = 'medium_purple4'
CC = 'cc:'
MIN_AUTHOR_PANEL_WIDTH = 80
EMAILER_INFO_TITLE = 'Email Conversations Will Appear'
UNINTERESTING_CC_INFO = "cc: or bcc: recipient only"
UNINTERESTING_CC_INFO_NO_CONTACT = f"{UNINTERESTING_CC_INFO}, no direct contact with Epstein"

INVALID_FOR_EPSTEIN_WEB = JUNK_EMAILERS + MAILING_LISTS + [
    'ACT for America',
    'BS Stern',
    UNKNOWN,
]


@dataclass(kw_only=True)
class Person:
    """Collection of data about someone texting or emailing Epstein."""
    name: Name
    emails: list[Email] = field(default_factory=list)
    imessage_logs: list[MessengerLog] = field(default_factory=list)
    other_files: list[OtherFile] = field(default_factory=list)
    is_uninteresting: bool = False

    def __post_init__(self):
        self.emails = Document.sort_by_timestamp(self.emails)
        self.imessage_logs = Document.sort_by_timestamp(self.imessage_logs)

    @property
    def category(self) -> str | None:
        highlight_group = self.highlight_group

        if highlight_group and isinstance(highlight_group, HighlightedNames):
            category = highlight_group.category or highlight_group.label

            if category != self.name and category != 'paula':  # TODO: this sucks
                return category

    @property
    def category_txt(self) -> Text | None:
        if self.name is None:
            return None
        elif self.category:
            return styled_category(self.category)
        elif self.is_a_mystery or self.is_uninteresting:
            return QUESTION_MARKS_TXT

    @property
    def email_conversation_length_in_days(self) -> int:
        return days_between(self.emails[0].timestamp, self.emails[-1].timestamp)

    @property
    def earliest_email_at(self) -> datetime:
        return self.emails[0].timestamp

    @property
    def earliest_email_date(self) -> date:
        return self.earliest_email_at.date()

    @property
    def last_email_at(self) -> datetime:
        return self.emails[-1].timestamp

    @property
    def last_email_date(self) -> date:
        return self.last_email_at.date()

    @property
    def emails_by(self) -> list[Email]:
        return [e for e in self.emails if self.name == e.author]

    @property
    def emails_to(self) -> list[Email]:
        return [
            e for e in self.emails
            if self.name in e.recipients or (self.name is None and len(e.recipients) == 0)
        ]

    @property
    def external_links_line(self) -> Text:
        links = [self.external_link_txt(site) for site in PERSON_LINK_BUILDERS]
        return Text('', justify='center', style='dim').append(join_texts(links, join=' / '))  #, encloser='()'))#, encloser='‹›'))

    @property
    def has_any_epstein_emails(self) -> bool:
        contacts = [e.author for e in self.emails] + flatten([e.recipients for e in self.emails])
        return JEFFREY_EPSTEIN in contacts

    @property
    def highlight_group(self) -> HighlightedNames | None:
        return get_highlight_group_for_name(self.name)

    @property
    def info_panel(self) -> Padding:
        """Return a `Panel` with the name of an emailer and a few tidbits of information about them."""
        style = 'white' if (not self.style() or self.style() == DEFAULT) else self.style()
        panel_style = f"black on {style} bold"

        if self.name == JEFFREY_EPSTEIN:
            email_count = len(self._printable_emails())
            title_suffix = f"sent by {JEFFREY_EPSTEIN} to himself"
        else:
            email_count = len(self.unique_emails)
            num_days = self.email_conversation_length_in_days
            title_suffix = f"{TO_FROM} {self.name_str} starting {self.earliest_email_date} covering {num_days:,} days"

        title = f"Found {email_count} emails {title_suffix}"
        width = max(MIN_AUTHOR_PANEL_WIDTH, len(title) + 4, len(self.info_with_category) + 8)
        panel = Panel(Text(title, justify='center'), width=width, style=panel_style)
        elements: list[RenderableType] = [panel]

        if self.info_with_category:
            elements.append(Text(f"({self.info_with_category})", justify='center', style=f"{style} italic"))

        return Padding(Group(*elements), (2, 0, 1, 0))

    @property
    def info_str(self) -> str | None:
        highlight_group = self.highlight_group

        if highlight_group and isinstance(highlight_group, HighlightedNames) and self.name:
            info = highlight_group.info_for(self.name)

            if info:
                return info

        if self.is_uninteresting and len(self.emails_by) == 0:
            if self.has_any_epstein_emails:
                return UNINTERESTING_CC_INFO
            else:
                return UNINTERESTING_CC_INFO_NO_CONTACT

    @property
    def info_txt(self) -> Text | None:
        if self.name == JEFFREY_EPSTEIN:
            return Text('(emails sent by Epstein to himself are here)', style=ALT_INFO_STYLE)
        elif self.name is None:
            return Text('(emails whose author or recipient could not be determined)', style=ALT_INFO_STYLE)
        elif self.category == JUNK:
            return Text(f"({JUNK} mail)", style='bright_black dim')
        elif self.is_uninteresting and (self.info_str or '').startswith(UNINTERESTING_CC_INFO):
            if self.sole_cc:
                return Text(f"(cc: from {self.sole_cc} only)", style='wheat4 dim')
            elif self.info_str == UNINTERESTING_CC_INFO:
                return Text(f"({self.info_str})", style='wheat4 dim')
            else:
                return Text(f"({self.info_str})", style='plum4 dim')
        elif self.is_a_mystery:
            return Text(QUESTION_MARKS, style='honeydew2 bold')
        elif self.info_str is None:
            if self.name in MAILING_LISTS:
                return Text('(mailing list)', style=f"pale_turquoise4 dim")
            elif self.category:
                return Text(QUESTION_MARKS, style=self.style())
            else:
                return None
        else:
            return Text(self.info_str, style=self.style(allow_bold=False))

    @property
    def info_with_category(self) -> str:
        return ', '.join(without_falsey([self.category, self.info_str]))

    @property
    def internal_link(self) -> Text:
        """Kind of like an anchor link to the section of the page containing these emails."""
        return link_text_obj(internal_link_to_emails(self.name_str), self.name_str, style=self.style())

    @property
    def is_a_mystery(self) -> bool:
        """Return True if this is someone we theroetically could know more about."""
        return self.is_unstyled and not (self.is_email_address or self.info_str or self.is_uninteresting)

    @property
    def is_email_address(self) -> bool:
        return '@' in (self.name or '')

    @property
    def is_linkable(self) -> bool:
        """Return True if it's likely that EpsteinWeb has a page for this name."""
        if self.name is None or ' ' not in self.name:
            return False
        elif self.is_email_address or '/' in self.name or QUESTION_MARKS in self.name:
            return False
        elif self.name in INVALID_FOR_EPSTEIN_WEB:
            return False

        return True

    @property
    def is_unstyled(self) -> bool:
        """True if there's no highlight group for this name."""
        return self.style() == DEFAULT_NAME_STYLE

    @property
    def name_link(self) -> Text:
        """Will only link if it's worth linking, otherwise just a Text object."""
        if not self.is_linkable:
            return self.name_txt
        else:
            return Text.from_markup(link_markup(self.external_link(), self.name_str, self.style()))

    @property
    def name_str(self) -> str:
        return self.name or UNKNOWN

    @property
    def name_txt(self) -> Text:
        return styled_name(self.name)

    @property  # TODO: unused?
    def should_always_truncate(self) -> bool:
        """True if we want to truncate all emails to/from this user."""
        return self.name in TRUNCATE_EMAILS_FROM or self.is_uninteresting

    @property
    def sole_cc(self) -> str | None:
        """Return name if this person sent 0 emails and received CC from only one that name."""
        email_authors = uniquify([e.author for e in self.emails_to])

        if len(self.unique_emails) == 1 and len(email_authors) > 0:
            logger.info(f"sole author of email to '{self.name}' is '{email_authors[0]}'")
        else:
            logger.info(f"'{self.name}' email_authors '{email_authors[0]}'")

        if len(self.unique_emails_by) > 0:
            return None

        if len(email_authors) == 1:
            return email_authors[0]

    @property
    def sort_key(self) -> list[int | str]:
        """Key used to sort `Person` objects by the number of emails sent/received."""
        counts = [
            len(self.unique_emails),
            -1 * int((self.info_str or '') == UNINTERESTING_CC_INFO_NO_CONTACT),
            -1 * int((self.info_str or '') == UNINTERESTING_CC_INFO),
            int(self.has_any_epstein_emails),
        ]

        counts = [-1 * count for count in counts]

        if args.sort_alphabetical:
            return [self.name_str] + counts
        else:
            return counts + [self.name_str]

    @property
    def unique_emails(self) -> Sequence[Email]:
        return Document.without_dupes(self.emails)

    @property
    def unique_emails_by(self) -> list[Email]:
        return Document.without_dupes(self.emails_by)

    @property
    def unique_emails_to(self) -> list[Email]:
        return Document.without_dupes(self.emails_to)

    def external_link(self, site: ExternalSite = EPSTEINIFY) -> str:
        return PERSON_LINK_BUILDERS[site](self.name_str)

    def external_link_txt(self, site: ExternalSite = EPSTEINIFY, link_str: str | None = None) -> Text:
        if self.name is None:
            return Text('')

        return link_text_obj(self.external_link(site), link_str or site, style=self.style())

    def print_emails(self) -> list[Email]:
        """Print complete emails to or from a particular 'author'. Returns the Emails that were printed."""
        print_centered(self.info_panel)
        self.print_emails_table()
        last_printed_email_was_duplicate = False

        if self.category == JUNK:
            logger.warning(f"Not printing junk emailer '{self.name}'")
        else:
            for email in self._printable_emails():
                if email.is_duplicate:
                    console.print(Padding(email.duplicate_file_txt.append('...'), SKIPPED_FILE_MSG_PADDING))
                    last_printed_email_was_duplicate = True
                else:
                    if last_printed_email_was_duplicate:
                        console.line()

                    console.print(email)
                    last_printed_email_was_duplicate = False

        return self._printable_emails()

    def print_emails_table(self) -> None:
        table = Email.build_emails_table(self._unique_printable_emails(), self.name)
        print_centered(Padding(table, (0, 5, 0, 5)))

        if self.is_linkable:
            print_centered(self.external_links_line)

        console.line()

    def style(self, allow_bold: bool = True) -> str:
        return get_style_for_name(self.name, allow_bold=allow_bold)

    def _printable_emails(self):
        """For Epstein we only want to print emails he sent to himself."""
        if self.name == JEFFREY_EPSTEIN:
            return [e for e in self.emails if e.is_note_to_self]
        else:
            return self.emails

    def _unique_printable_emails(self):
        return Document.without_dupes(self._printable_emails())

    def __str__(self):
        return f"{self.name_str}"

    @staticmethod
    def emailer_info_table(people: list['Person'], highlighted: list['Person'] | None = None, show_epstein_total: bool = False) -> Table:
        """Table of info about emailers."""
        highlighted = highlighted or people
        highlighted_names = [p.name for p in highlighted]
        is_selection = len(people) != len(highlighted) or args.emailers_info
        all_emails = Person.emails_from_people(people)
        email_authors = [p for p in people if p.emails_by and p.name]
        attributed_emails = [email for email in all_emails if email.author]
        footer = f"(identified {len(email_authors)} authors of {len(attributed_emails):,}" \
                 f" out of {len(all_emails):,} emails, {len(all_emails) - len(attributed_emails)} still unknown)"

        if is_selection:
            title = Text(f"{EMAILER_INFO_TITLE} in This Order for the Highlighted Names (", style=TABLE_TITLE_STYLE)
            title.append(THE_OTHER_PAGE_TXT).append(" has the rest)")
        else:
            title = f"{EMAILER_INFO_TITLE} in Chronological Order Based on Timestamp of First Email"

        table = build_table(title, caption=footer)
        table.add_column('First')
        table.add_column('Name', max_width=24, no_wrap=True)
        table.add_column('Category', justify='left', style='dim italic')
        table.add_column('Num', justify='right', style='white')
        table.add_column('Sent', justify='right', style='wheat4')
        table.add_column('Recv', justify='right', style='wheat4')
        table.add_column('Days', justify='right', style=TIMESTAMP_DIM)
        table.add_column('Info', style='white italic')
        current_year = 1990
        current_year_month = current_year * 12
        grey_idx = 0

        for person in people:
            earliest_email_date = person.earliest_email_date
            is_on_page = False if show_epstein_total else person.name in highlighted_names
            year_months = (earliest_email_date.year * 12) + earliest_email_date.month

            # Color year rollovers more brightly
            if current_year != earliest_email_date.year:
                grey_idx = 0
            elif current_year_month != year_months:
                grey_idx = ((current_year_month - 1) % 12) + 1

            current_year_month = year_months
            current_year = earliest_email_date.year

            table.add_row(
                Text(str(earliest_email_date), style=f"grey{GREY_NUMBERS[0 if is_selection else grey_idx]}"),
                person.internal_link if is_on_page and not person.is_uninteresting else person.name_txt,
                person.category_txt,
                f"{len(person.unique_emails if show_epstein_total else person._unique_printable_emails())}",
                str(len(person.unique_emails_by)) if len(person.unique_emails_by) > 0 else '',
                str(len(person.unique_emails_to)) if len(person.unique_emails_to) > 0 else '',
                f"{person.email_conversation_length_in_days}",
                person.info_txt or '',
                style='' if show_epstein_total or is_on_page else 'dim',
            )

        return table

    @staticmethod
    def emails_from_people(people: list['Person']) -> Sequence[Email]:
        return Document.uniquify(flatten([list(p.unique_emails) for p in people]))



"""
Regexes and patterns for identifying people in email headers.
"""
import re
from copy import deepcopy

from epstein_files.util.constant.names import *
from epstein_files.util.constant.strings import REDACTED
from epstein_files.util.data import escape_single_quotes
from epstein_files.util.logging import logger

BAD_EMAILER_REGEX = re.compile(r'^(>|11111111)|agreed|ok|sexy|re:|fwd:|Multiple Senders|((sent|attachments|subject|importance).*|.*(january|201\d|hysterical|i have|image0|so that people|article 1.?|momminnemummin|These conspiracy theories|your state|undisclosed|www\.theguardian|talk in|it was a|what do|cc:|call (back|me)|afiaata|[IM]{4,}).*)$', re.IGNORECASE)
BAD_NAME_CHARS_REGEX = re.compile(r"[\"'\[\]*><•=()]")
TIME_REGEX = re.compile(r'^((\d{1,2}/\d{1,2}/\d{2,4}|Thursday|Monday|Tuesday|Wednesday|Friday|Saturday|Sunday)|\d{4} ).*')

EMAILER_ID_PATTERNS: dict[str, str] = {
    ALAN_DERSHOWITZ: r'(alan.{1,7})?dershowi(lz?|t?z)|AlanDersh',
    ALIREZA_ITTIHADIEH: r'Alireza.[Il]ttihadieh',
    ALISON_J_NATHAN: r"Alison(\s*J\.?)?\s*Nathan|Nathan NYSD Chambers?",
    AMANDA_ENS: r'ens, amanda?|Amanda.Ens',
    AMIR_TAAKI: r'Amir\s*Taaki|genjix',
    ANAS_ALRASHEED: r'anas\s*al\s*rashee[cd]',
    ANIL_AMBANI: r'Anil.Ambani',
    ANN_MARIE_VILLAFANA: r'Villafana, Ann Marie|(A(\.|nn) Marie )?Villafa(c|n|ri)a',
    ANTHONY_SCARAMUCCI: r"mooch|(Anthony ('The Mooch' )?)?Scaramucci",
    ARIANE_DE_ROTHSCHILD: r'AdeR|((Ariane|Edmond) (de )?)?Rothsh?ch?ild|Ariane(?!\s+Dwyer)',
    BARBRO_C_EHNBOM: r'behnbom@aol.com|(Barbro\s.*)?Ehnbom',
    BARRY_J_COHEN: r'barry\s*((j.?|james)\s*)?cohen?',
    BENNET_MOSKOWITZ: r'Moskowitz.*Bennet|Bennet.*Moskowitz',
    BOB_CROWE: r"[BR]ob Crowe",
    BORIS_NIKOLIC: r'(boris )?nikolic?',
    BRAD_EDWARDS: r'Brad(ley)?(\s*J(.?|ames))?\s*Edwards',
    BRAD_KARP: r'Brad (S.? )?Karp|Karp, Brad',
    CHRISTIAN_EVERDELL: r"C(hristian\s*)?Everdell?",
    CHRISTOPHER_DILORIO: r"Chris\s*Di[lI]o[nr](io)?",
    DANGENE_AND_JENNIE_ENTERPRISE: r'Dangene and Jennie Enterprise?',
    DANNY_FROST: r'Frost, Danny|frostd@dany.nyc.gov|Danny\s*Frost',
    DARREN_INDYKE: r'darren$|Darren\s*(K\.?\s*)?[il]n[dq]_?yke?|dkiesq',
    DAVID_FISZEL: r'David\s*Fis?zel',
    DAVID_HAIG: fr'{DAVID_HAIG}|Haig, David',
    DAVID_STERN: r'David Stern?',
    DOUGLAS_WIGDOR: r'Doug(las)?\s*(H\.?)?\s*Wigdor',
    EDUARDO_ROBLES: r'Ed(uardo)?\s*Robles',
    EDWARD_JAY_EPSTEIN: r'(?<!Jeffrey )Edward (Jay )?Epstein',
    EHUD_BARAK: r'(ehud|e?h)\s*barak|\behud',
    FAITH_KATES: r'faith kates?',
    GERALD_BARTON: r'Gerald.*Barton',
    GERALD_LEFCOURT: r'Gerald\s*(B\.?\s*)?Lefcourt',
    GHISLAINE_MAXWELL: r'g ?max(well)?|Ghislaine|Maxwell',
    HEATHER_MANN: r'Heather Mann?',
    INTELLIGENCE_SQUARED: r'intelligence\s*squared',
    JACKIE_PERCZEK: r'jackie percze[kl]?',
    JABOR_Y: r'[ji]abor\s*y?',
    JAMES_HILL: r"hill, james e.|james.e.hill@abc.com",
    JANUSZ_BANASIAK: r"Janu[is]z Banasiak",
    JEAN_HUGUEN: r"Jean[\s.]Huguen",
    JEAN_LUC_BRUNEL: r'Jean[- ]Luc Brunel?|JeanLuc',
    JEFF_FULLER: r"jeff@mc2mm.com|Jeff Fuller",
    JEFFREY_EPSTEIN: r'[djl]\s?ee[vy]acation[©@]?g?(mail.com)?|Epstine|\bJEE?\b|Jeff(rey)? (Edward )?E((sp|ps)tein?)?( VI Foundation)?|jeeproject@yahoo.com|J Jep|Jeffery Edwards|(?<!(Mark L.|ard Jay) )Epstein',
    JESSICA_CADWELL: r'Jessica Cadwell?',
    JOHNNY_EL_HACHEM: r'el hachem johnny|johnny el hachem',
    JOI_ITO: r'ji@media.mit.?edu|(joichi|joi)( Ito)?',
    JONATHAN_FARKAS: r'Jonathan Fark(a|u)(s|il)',
    KARYNA_SHULIAK: r"Karyna\s*Shuliak?",
    KATHRYN_RUEMMLER: r'Kathr?yn? Ruemmler?',
    KEN_STARR: r'starr, ken|Ken(neth\s*(W.\s*)?)?\s+starr?|starr',
    LANDON_THOMAS: r'lando[nr] thomas( jr)?|thomas jr.?, lando[nr]',
    LARRY_SUMMERS: r'(La(wrence|rry).{1,5})?Summers?|^LH$|LHS|[Il]hsofficel?',
    LAWRANCE_VISOSKI: r'La(rry|wrance) Visoski?|Lvjet',
    LAWRENCE_KRAUSS: r'Lawrence Kraus[es]?|[jl]awkrauss|kruase',
    LEON_BLACK: r'Leon\s*Black?|(?<!Marc )Leon(?! (Botstein|Jaworski|Wieseltier))',
    LILLY_SANCHEZ: r'Lilly.*Sanchez',
    LISA_NEW: r'E?Lisa New?\b',
    MANUELA_MARTINEZ: fr'Manuela (- Mega Partners|Martinez)',
    MARIANA_IDZKOWSKA: r'Mariana [Il]d[źi]kowska?',
    MARK_EPSTEIN: r'Mark (L\. )?(Epstein|Lloyd)',
    MARC_LEON: r'Marc[.\s]+(Kensington|Leon)|Kensington2',
    MARTIN_NOWAK: r'(Martin.*?)?No[vw]ak|Nowak, Martin',
    MARTIN_WEINBERG: r'martin.*?weinberg',
    "Matthew Schafer": r"matthew\.?schafer?",
    MELANIE_SPINELLA: r'M?elanie Spine[Il]{2}a',
    MICHAEL_BUCHHOLTZ: r'Michael.*Buchholtz',
    MICHAEL_MILLER: r'Micha(el)? Miller|Miller, Micha(el)?',
    MICHAEL_SITRICK: r'(Mi(chael|ke).{0,5})?[CS]itrick',
    MICHAEL_WOLFF: r'Michael\s*Wol(f[ef]e?|i)|Wolff',
    MIROSLAV_LAJCAK: r"Miro(slav)?(\s+Laj[cč][aá]k)?",
    MOHAMED_WAHEED_HASSAN: r'Mohamed Waheed(\s+Hassan)?',
    NADIA_MARCINKO: r"Na[dď]i?a\s+Marcinko(v[aá])?",
    NEAL_KASSELL: r'Neal\s*Kassell?',
    NICHOLAS_RIBIS: r'Nic(holas|k)[\s._]Ribi?s?|Ribbis',
    OLIVIER_COLOM: fr'Colom, Olivier|{OLIVIER_COLOM}',
    PAUL_BARRETT: r'Paul Barre(d|tt)',
    PAUL_KRASSNER: r'Pa\s?ul Krassner',
    PAUL_MORRIS: r'morris, paul|Paul Morris',
    PAULA: r'^Paula( Heil Fisher)?$',
    PEGGY_SIEGAL: r'Peggy Siegal?',
    PETER_ATTIA: r'Peter Attia?',
    PETER_MANDELSON: r"((Lord|Peter) )?Mandelson",
    'pink@mc2mm.com': r"^Pink$|pink@mc2mm\.com",
    PRINCE_ANDREW: r'Prince Andrew|The Duke',
    REID_WEINGARTEN: r'Weingarten, Rei[cdi]|Rei[cdi] Weingarten',
    RICHARD_KAHN: r'rich(ard)? kahn?',
    ROBERT_D_CRITTON_JR: r'Robert D.? Critton,? Jr.?',
    ROBERT_LAWRENCE_KUHN: r'Robert\s*(Lawrence)?\s*Kuhn',
    ROBERT_TRIVERS: r'tri[vy]ersr@gmail|Robert\s*Trivers?',
    ROSS_GOW: fr"Ross(acuity)? Gow|(ross@)?acuity\s*reputation(\.com)?",
    SAMUEL_LEFF: r"Sam(uel)?(/Walli)? Leff",
    SCOTT_J_LINK: r'scott j. link?',
    SEAN_BANNON: r'sean bannon?',
    SHAHER_ABDULHAK_BESHER: r'\bShaher( Abdulhak Besher)?\b',
    SOON_YI_PREVIN: r'Soon[- ]Yi Previn?',
    STACEY_RICHMAN: r"srichmanlaw|Stacey\s*Richman",
    STEPHEN_HANSON: r'ste(phen|ve) hanson?|Shanson900',
    STEVE_BANNON: r'steve banno[nr]?',
    STEVEN_SINOFSKY: r'Steven Sinofsky?',
    SULTAN_BIN_SULAYEM: r'Sultan (Ahmed )?bin Sulaye?m?',
    TERJE_ROD_LARSEN: r"Terje(( (R[øo]e?d[- ])?)?Lars[eo]n)?",
    TERRY_KAFKA: r'Terry Kafka?',
    THANU_BOONYAWATANA: r"Thanu (BOONYAWATANA|Cnx)",
    THORBJORN_JAGLAND: r'(Thor.{3,8})?Jag[il]and?',
    TONJA_HADDAD_COLEMAN: r"To(nj|rl)a Haddad Coleman|haddadfm@aol.com",
    VINCENZO_IOZZO: r"Vincenzo [IL]ozzo",
}

# If found as substring consider them the author
EMAILERS = [
    'Anne Boyles',
    AL_SECKEL,
    'Ariane Dwyer',
    AZIZA_ALAHMADI,
    BILL_GATES,
    BILL_SIEGEL,
    'Bobbi C Sternheim',
    BRAD_WECHSLER,
    BROCK_PIERCE,
    CHRISTINA_GALBRAITH,
    DANIEL_SABBA,
    'Danny Goldberg',
    DAVID_SCHOEN,
    DEBBIE_FEIN,
    DEEPAK_CHOPRA,
    GLENN_DUBIN,
    GORDON_GETTY,
    'Jeff Pagliuca',
    'Kevin Bright',
    'Jack Lang',
    JACK_SCAROLA,
    JAY_LEFKOWITZ,
    JES_STALEY,
    JOHN_PAGE,
    'Jokeland',
    JOSCHA_BACH,
    'Kathleen Ruderman',
    KENNETH_E_MAPP,
    'Larry Cohen',
    LESLEY_GROFF,
    'lorraine@mc2mm.com',
    LINDA_STONE,
    'Lyn Fontanilla',
    MARK_TRAMO,
    MELANIE_WALKER,
    MERWIN_DELA_CRUZ,
    'Michael Simmons',   # Not the only "To:"
    'middle.east.update@hotmail.com',
    'Nancy Cain',
    'Nancy Dahl',
    'Nancy Portland',
    'Nathan NYSD Chambers',
    'Oliver Goodenough',
    'Paula Speer',
    'Peter Aldhous',
    'Peter Green',
    ROGER_SCHANK,
    'Roy Black',
    STEVEN_PFEIFFER,
    'Steven Victor MD',
    'Susan Edelman',
    TOM_BARRACK,
    'USANYS',
    'Vahe Stepanian',
    'Vladimir Yudashkin',
]

EMAILER_ID_REGEXES = {name: re.compile(pattern, re.IGNORECASE) for name, pattern in EMAILER_ID_PATTERNS.items()}
EMAILER_REGEXES = deepcopy(EMAILER_ID_REGEXES)  # Keep a copy without the simple EMAILERS regexes

# Add simple matching regexes for EMAILERS entries to EMAILER_REGEXES
for emailer in EMAILERS:
    if emailer in EMAILER_REGEXES:
        raise RuntimeError(f"Can't overwrite emailer regex for '{emailer}'")

    EMAILER_REGEXES[emailer] = re.compile(emailer + '?', re.IGNORECASE)  # Last char optional bc OCR sucks

SUPPRESS_LOGS_FOR_AUTHORS = [
    'Multiple Senders Multiple Senders',
    'Undisclosed recipients:',
    'undisclosed-recipients:',
]


def cleanup_str(_str: str) -> str:
    return BAD_NAME_CHARS_REGEX.sub('', _str.replace(REDACTED, '')).strip().strip('_').strip()


def extract_emailer_names(emailer_str: str) -> list[str]:
    """Return a list of people's names found in `emailer_str` (email author or recipients field)."""
    emailer_str = cleanup_str(emailer_str)

    if len(emailer_str) == 0:
        return []

    names_found = [name for name, regex in EMAILER_REGEXES.items() if regex.search(emailer_str)]

    if len(emailer_str) <= 2 or BAD_EMAILER_REGEX.match(emailer_str) or TIME_REGEX.match(emailer_str):
        if len(names_found) == 0 and emailer_str not in SUPPRESS_LOGS_FOR_AUTHORS:
            logger.warning(f"No emailer found in '{escape_single_quotes(emailer_str)}'")
        else:
            logger.info(f"Extracted {len(names_found)} names from semi-invalid '{emailer_str}': {names_found}...")

        return names_found

    names_found = names_found or [emailer_str]
    return [reverse_first_and_last_names(name) for name in names_found]



EFTA00009622_TEXT = """
[redacted] s GJ Test.          7/19/2006

[redacted] dob          16 yrs

[redacted] - Loxahatchee

2005 [redacted]          9th grade

[redacted] was outside          asked to date [redacted]

1/10/19 [redacted]

~~talked in depth~~

met for first time the night before went to JE
11:30 ish 10
do you want to make $200

met my friend Jeff  bighouse

but you have to give massage 45 min + that's all you have to do get $200

[redacted] told [redacted] + [redacted] went to Bathroom
to argue
came back out [redacted] called Jeff on the phone.
1 assistant answered the pho...

Page 2
Jeff asst answered phone - [redacted] had someone to give massage Lady asked how old I was
she told the lady 18 + went to [redacted]

Its $200     didn't care what
[redacted] - Lady trying to verify 18 Yes
Graduated from [redacted]     Known [redacted] for a long time
& the next morning

[redacted] + another girl     - going to Mall he [redacted] didn't care

[redacted] asked dad for $     yeh sure fine ok

House in Palm Beach
all got out of car     gate at back door     security guard the saw us     the [redacted] asked us where...
Page 3

[redacted] said Jeffrey
Went to kitchen     waited 15 min

JE + asst     walked in back door

shook hands     met both JE asst.
(A) JE     chose gonna go first

JE said how about you
[redacted]     upstairs     "asst lady"

walked me up there stripping out all this picture

She pulled out massage table me     front of couch     blanket over table     opened drawer     whole bunch of lotions     + pick some out
She told me keep bra + panties on + get undress

No [redacted]     told her I thought a massage keep your clothing

Page 4

stay in the room
I came in the room     HE + shook hands
Came back in the room

JE:     old     50ish

initially     guy golfy     khaki pants pullover

t-shirt + jeans     thong     bra

hesitant     thinking what's going on. I didn't care cause I wanted $200

He laid on massage table     He said grab any 3 lotions

Started to give him

What M.S.     did you go to told     him she was 16 Not     do you like to do     small talk [redacted]

Sat on table     15-20 minute conversation     another house in NY

Page 5

He asked if I wanted to make an extra $300
Straddle him

massage over -
but if you're not comfortable giving a massage can I     [redacted]     on you     ~~anything~~
[redacted]     said ok
JE left + came ba

[redacted]     kind of like a [redacted]
grabbed that + layed down back on table     ~~old~~ ~~old~~ hands

Give me a     massage on chest use the [redacted]     10 minutes or so ( after     a couple ~~of~~ minutes ( on [redacted]
(A) [redacted] said if he did that he would have to pay her extra money if he >
Page 6

[redacted]     not she would have to didn't     tell her     ~~get~~ ~~over~~ do her clothes at He [redacted]     you get paid

[redacted]

Yes     Saw [redacted]

He opened wallet     got the   $200

He told her to write Name + #

after that left the room

I didn't see him [redacted]

I was scared to tell

[redacted]     her body not his body
Went down by herself    got 1st  No didn't do anything  but [redacted]  $200 ~~too~~

Page 7
I figure it was probably because she brought me

What happened     I got $300 I was excited

What did you have to do with him
If you do this you get that
Laughing ~~in~~     Thanks for fully me I was going to ~~much~~ ~~go~~ ~~many~~ Oh sorry.
Didn't plan it

If you ever need money
Oh yeh     - everybody

I did not ever go back to house

at school my best friend I told her if you ever want to make 300.-o massage Told [redacted]     [redacted]

Page 8
[redacted]     telling everybody I confronted her     altercation get into a fight [redacted]

Spreading rumors about me Let me see in your purse you don't have to lie Stefand I worked at [redacted]

I didn't think it was her business and     why would tell on myself

You don't have to lie I knew what happen
Everything came out

[redacted]     then I didn't want to tell them what had happened,

I [redacted]

[redacted] Family Setty a lot problems

No     the only [redacted]     sometimes
Page 9

[redacted]     Have you ever da[ting] [redacted] Survey -     Yeh
[redacted]     not anymore
14 birthday     my dad
I did [redacted]     lots of time

$250,000     its a joke
tall skinny     dark hair
Page 10
""".strip()
